/*****************************************************************************/
/*
                                  mondesi.c

                        Monitor de Sistema (espaol)

MONDESI (pronounced "mon-deh-sih") is a browser-based VMS system monitor with a
sufficiently small resource footprint as not to impact system performance
(minimal CPU consumption and only one buffered IO per client update).  No, the
name's etiology has little significance at all (a small homage at most).  I was
just looking for a short (and cute) name.

This utility is intended as a general tool for system observation.  It is not a
troubleshooting tool as such, especially when a system is beginning to suffer
from resource exhaustion.  Not only does the data collection script need to
continue processing effectively (see 'SUFFICIENT PRIORITY' below) there is also
a Web server in the data transfer path.

MONDESI should work with all relatively modern browsers (the graphics toolkit -
see below - is the primary limitation); those with the Gecko-Engine - Mozilla,
SeaMonkey, Netscape 6+, Firefox, along with Opera 7+, and MSIE 5+.  The intense
graphing is CPU and memory hungry so I'd not recommend Mozilla on an DEC 3000
(or even a PWS 500 :-)  Of course YMMV.

MONDESI should also work with HTML5 WebSocket capable browsers.

Uses the system service $GETRMI() available for VMS V7.3 and later.
Note that V7.3-1 (and earlier) do not support some data items.  These are shown
as question marks.

Displayed values have been validated using the interactive MONITOR utility.

As much of the functionality is pushed off to the browser MONDESI relies
heavily on JavaScript contained in MONDESI.JS and WS_JSGRAPHICS.JS (see below).

As MONDESI provides a readily accessable insight into a system (albeit,
basically resource consumption information only) perhaps it would be best
available only when under authorization.

NOTE: MONDESI was essentially a vehicle for experimentation rather than an
application designed and built to address a specific need.  On the other hand
it may be useful to someone.  Enjoy!


MONITOR DISPLAY
---------------
A node is displayed as a text panel on the left and nine bar graphs on the
right.  These display the instantaneous (current) values for the various
resources and are updated dynamically at the interval when the data collector
script detects a change in values.  Each bar graph section contains a number
representing the current value and a number in the far right the maximum value
reached while monitoring.  All values are per-second where applicable (e.g.
BIO, DIO, lock, etc.)

  +--------------------------------------------------------------------------+
  | NODENAME                   CPU:[_______________________:___|___________] |
  | 21-SEP-2010 21:44          MEM:[______________|________________________] |
  | Digital Personal Workstn  FLTS:[_______________________________________] |
  | 1 CPU 766 MB Alpha V8.3    BIO:[_____:_|_______________________________] |
  | Up: 29 05:50 Processes:25  DIO:[________________|______________________] |
  | interactive:3 batch:0     LOCK:[_|_____________________________________] |
  | network:11 (v2.0.0/WASD)    NI:[______________________________|________] |
  |                           DISK:[__:_________|__________________________] |
  | [hi][cl][pr][re][ze][st]   PRC:[_|_____________________________________] |
  +--------------------------------------------------------------------------+

CPU: shows total CPU utilisation in green, non-user-mode CPU
(kernel+interrupt+etc.) in brown (always a subset of total) and average CPU
utilisation in yellow.  Where a system has single CPU the graph range is 100%,
with two CPUs the range is 200%, three is 300%, etc.

MEM: displays system physical memory (RAM) in Mbytes.  The value in the far
right is the total.  The current used is show in green.  Used memory is
calculate by subtracting the current free list plus modified list sizes from
the total physical memory.

FLTS: provides total page faults per second in green with hard page faults as a
subset in brown.  Average page faulting for the monitor period is shown in
yellow.

BIO: displays buffered I/Os per second.  Average is shown in yellow.

DIO: displays direct I/Os per second.  Average is shown in yellow.

LOCK: provides locking activity as a total of ENQs plus DEQs plus waits. 
Locks waits are shown in brown.  Average activity is shown in yellow.

NI: network interface(s) bytes received and transmitted per second; total in
green, received in brown, transmitted data being the difference.  This includes
all data moving through the physical network interface(s) including SCS, MSCP,
etc., not just protocols such as TCP/IP. 

DISK: displays total disk I/Os in green with a subset of disk writes in
brown (disk reads being the difference).  Average activity shown in yellow.

[history] button reveals/hides the history line graphs (see below).

[cluster] enabled if a clustered node, button reveals/hides cluster data

[print] suspends updates to allow use of the print dialog

[scale] adjusts the y axis to the current maximum data value

[zero] button can be used to reset the collected maximum and average data.

[stop][start] stops data collection while maintaining the display
              start (re)starts data collection and zeroes display

When cluster button reveals cluster data the following additional bar graphs
and associated history graphs become visible.

  +--------------------------------------------------------------------------+
  :                                                                          :
  |                          SCSMG:[______________________________|________] |
  |                          SCSKB:[__:_____________|______________________] |
  | [hi][cl][pr][re][ze][st]  MSCP:[_|_____________________________________] |
  +--------------------------------------------------------------------------+

SCSMG: count of SCS messages exchanged.  Displays total in green and received
in brown.  (MONITOR> SCS /ITEM=(M_SEND,M_RECEIVE) analogue.)  Average is
displayed in yellow. 

SCSKB: kilobyte map rate.  Displays total in green.  (MONITOR> SCS
/ITEM=(KB_MAP) analogue.)  Average is displayed in yellow. 

MSCP: displays total requests from remote systems in green and a subset showing
write requests in brown.  Average is displayed in yellow.


NETWORK STATISTICS
------------------
MONDESI will scan the system for network devices (commonly ethernet) and use
any found for NI statistics.  If the system admin wishes to tailor the source
of the statistics the multi-valued logical name MONDESI_NI can be used to
specify one or more network interfaces.  Define this logical name system-wide
(or in a script wrapper procedure) as required.

For example:

  $ DEFINE /SYSTEM /EXECUTIVE MONDESI_NI EWA0:,EWB0:

This logical name must exist as MONDESI initialises.  If undefined or the
specified device(s) cannot be accessed the ethernet data are greyed-out.


HISTORY DISPLAY
---------------
Recent data are stored within the browser and used to generate line graphs
showing values against time.  These follow the same colouring schema as the
instantaneous bar graphs.   For example, the CPU consumption line is in green
and the non-user-mode in brown.

The [history] button reveals and hides the history graphs.  The complexity of
rendering these graphics puts a noticable load on the browser system.


PRINTING
--------
Printing a 'page' that is currently building itself (as these displays are
because of the hidden IFRAME receiving the resource data) is problematic on all
browsers tested.   As a workaround the data feed is closed (by removing the
hidden IFRAME) thus stopping the page building.  The JavaScript print is then
requested.  After a second or two the IFRAME is recreated resuming the data
feed and update of the displays.  Bit kludgy, but hey, it's a browser.  Firefox
and MSIE printed OK but Opera provided an unrendered page.  C'est la vie!

The [print] button is highlighted when first clicked.  Display updates (but not
data gathering) are suspended while the [print] button is engaged allowing a
static 'snapshot' of the data to be printed.  The browser print dialog can then
be used or cancelled.  Clicking the [print] button again once the print is
queued (and you should!) resumes display updates.  The [print] button remains
enabled for a maximum of 45 seconds before being automatically reset.


CONFIGURATION
-------------
The configuration page provides a text entry area into which URLs for the nodes
to monitor are entered, one per line.  The 'nodes across window' selector allow
monitor displays to be tiled (if there is enough screen real estate).  Using
this approach widely 'separated' hosts can be monitored on a single page.  The
URL generated by the configuration dialog can be saved as a bookmark, entered
into an HTML page link, etc., for ease of subsequent access.


SUFFICIENT PRIORITY
-------------------
For a monitor such as this to be effective on a busy system it needs to have a
bit of an edge!  If INSTALLed with ALTPRI it will bump it's process priority up
to 6 when starting and reduce it back to previous when exiting (just in case
it's WASD and a persistent scripting process :-) 

  $ INSTALL REPLACE <location>MONDESI.EXE /PRIV=ALTPRI

Don't forget to reINSTALL any subsequently built or deployed executable.


CONTROLLING ACCESS
------------------
MONDESI provides an insight into the system and therefore is a security
consideration.  The information is fairly benign but for this and other reasons
it is often desirable to control who may use it.  Of course this can be
exercised using the mapping and/or authorisation functionality of the
underlying web server.  This is a WASD example that merely makes MONDESI
unknown outside of the local subnet.

  # WASD_CONFIG_MAP (HTTPD$MAP)
  if (!remote-addr:192.168.1.0/24) pass /cgi-bin/mondesi* "404 Not Found"


WEBSOCKETS
----------
MONDESI v2.n optionally uses HTML5 WebSockets

  http://en.wikipedia.org/wiki/Websockets

to perform the data transfer from server to client.  The WASD server v10.1 and
later supports scripting using WebSockets as the IPC.  A single CGIplus
scripting process can support multiple WebSocket clients making it particularly
efficient.  The code must support multiple asynchonous threads of processing. 
MONDESI provides this using ASTs.  The WebSock..() set of funtions handle the
concurrent WebSocket processing.  See the WASD documentation for detailed
information on WebSocket support.

MONDESI detects the underlying request type and adjusts its behaviour between
standard HTTP clients (using the "Web Push" techology described below) and
WebSocket clients.

MONDESI uses only the simplest implementation of the WebSocket protocol.  It
expects all frames to be (non UTF-8) text beginning with 0x00 and ending in
0xff.  The close frame of 0xff followed by 0x00 is also noted.  Other frame
formats will cause the WebSocket to be closed.  All of the data exchanged on
the WebSocket is ASCII (0-127) and so no conversion to/from UTF-8 should be
required.


WEB "PUSH"
----------
MONDESI uses JavaScript and a Web 'push' technology (such as it is) to return
data (not images!) from a Web server scripting process to the client monitor
page where it is displayed as text or rendered into (crude but effective) bar
graphs.  Perhaps it would be better described as "HTTP streaming" or as "long
polling" but it *definitely* does not need to poll for each update.  The "push"
will last as long as the Web server and intermediate network connectivity
allows.

A parent page provides the JavaScript and HTML infrastructure to support
multiple monitored nodes each contained in an indepdendent, visible IFRAME,
which in turn used a dynamically created (i.e. JavaScript), hidden IFRAME to
execute the data collection URL (script).  This executes continuously,
periodically collecting data and outputting changed items into the response
stream.  Each response record contains a JavaScript tag element which contains
calls to the required data processing and display functions in the visible
IFRAME to process (only) each datum that needs updating.

If a data collection process (script) should exit for whatever reason (timeout,
network interruption, etc.) MONDESI should restart data collection
automatically.  If the monitored system become unavailable for any reason
(restart, network interruption, etc.) the monitor highlights the node
information with red after ninety seconds.  Elementary but suitable to the
desired outcome.


INTERFRAME COMMUNICATION (IFC)
------------------------------
The node(s) displayed in the parent page are not necessarily the same as the
parent page (indeed, multiple nodes could not all be) and of course these have
differing host names and/or IP addresses.  This incurs browser cross-domain
restrictions with communication between the node display HTML (i)frame and the
primary, top-level page, specifically - when needing to open/close the history
graph portion of the display associated with each node.

Two approaches for providing this IFC are used. 
Both are well documented on the Web.

1) The more elegant approach uses the modern, architected, HTML5 postMessage()
communication mechanism.  MONDESI checks for and uses this if available in the
browser.  It sends a message to the top-level page which invokes a function to
adjust the node display height.  Currently (at initial MONDESI development)
Firefox 3 (specifically Gecko 1.9), Opera 9.5 and MSIE 8 (at least) support it.

2) If postMessage() is unavailable MONDESI falls back to using an IFC *kludge*
involving the dynamic creation of an IFRAME within the node monitor IFRAME and
using this to load a URL for the served file BCHAN.HTML which contains a call
to a JavaScript function in the top-level page (to adjust the height).  The
data are contained in the hash portion of the URL and never appears outside the
browser.  This should be usable with most older, mainstream browsers.


GRAPHING
--------
This application uses a third-party JavaScript graphics toolkit to generate
both the bar and line graphs.

  Copyright (c) 2002-2009 Walter Zorn. All rights reserved.

  Performance optimizations for Internet Explorer
  by Thomas Frank and John Holdsworth.
  fillPolygon method implemented by Matthieu Haller.

  LICENSE: LGPL

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

Used to be documented at

  http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

Now accessable via WayBackMachine (split URL)

  http://replay.waybackmachine.org/
  20090302111911/http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

(and now archived in the [.JS] directory).

Many thanks to (the late) Walter for making this useful tool freely available. 
This toolkit works extraordinarily well considering it uses no graphics
capability in the browser!!  This is a non-SVG, non-VML, DOM kludge (not
meaning this as a pejorative).


BUILD DETAILS
-------------
See BUILD_MONDESI.COM procedure.

  $ @BUILD_MONDESI [LINK]


APACHE INSTALLATION
-------------------
Copy the executable to the APACHE$ROOT:[CGI-BIN] directory.

Map the JavaScript resources into web-space using (assumes the application is
located under APACHE$COMMON:[000000]):

  Alias /mondesi/-/ "/apache$common/mondesi/"


OSU ACCOMODATIONS
-----------------
OSU requires the use of the MONDESI.COM DCL wrapper to invoke the OSU CGI
environment utility before executing MONDESI.  The wrapper and the MONDESI.EXE
executable both should be placed in WWW_ROOT:[BIN].  Rather than introduce the
bulk of WASD CGILIB this deliberately slimline monitor utility checks for an
indication of the OSU environment and overtly handles the indiosyncracies.  
MONDESI does not reopen <stdout> in binary mode.  OSU requires two CGI response
headers (MONDESI supplies "Status:" and "Content-Type:").  OSU also dislikes a
response header with <CR><LF> carriage-control, so MONDESI uses only an <LF>.

Map the JavaScript resources into web-space using (assumes the application is
located under WWW_ROOT:[000000]):

  pass /mondesi/-/* /www_root/mondesi/*


WASD INSTALLATION
-----------------
Copy the executable to the CGI_EXE: directory.

Map the JavaScript resources into web-space using (assumes the application is
located in the source tree):

  # HTTPD$MAP
  pass /mondesi/-/* /ht_root/src/mondesi/*


COPYRIGHT
---------
Copyright (C) 2008-2010 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY (update SOFTWAREVN as well!)
---------------
23-OCT-2010  MGD  v2.0.0, happy birthday Raupu
                          add network interface (NI) data
                          add processes
                          add lock waits to LOCK display
                          add SCS messages
                          PGFL data is now shown by hovering over MEM
                          data are now presented as units, kilo, mega, giga
                          support WASD WebSocket environment
                          bugfix; nodes across
03-JUN-2009  MGD  v1.0.2, bumped the dot-point merely to identify
                          the 3.05 version of wz_jsgraphics.js
13-DEC-2008  MGD  v1.0.1, CollectRMI() allow for 7.3-1 and earlier (sigh!)
                          thanks Graham Burley for suggesting sys$gl_*jobcnt
29-NOV-2008  MGD  v1.0.0, initial
*/
/*****************************************************************************/

#define SOFTWAREVN "2.0.0"
#define SOFTWARENM "MONDESI"
#ifdef __ALPHA
   char SoftwareID [] = SOFTWARENM " AXP-" SOFTWAREVN;
#endif
#ifdef __ia64
   char SoftwareID [] = SOFTWARENM " IA64-" SOFTWAREVN;
#endif
#ifdef __VAX
   char SoftwareID [] = SOFTWARENM " VAX-" SOFTWAREVN;
#endif

/* standard C header files */
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* VMS-related header files */
#include <descrip.h>
#include <dvidef.h>
#include <dvsdef.h>
#include <iodef.h>
#include <jpidef.h>
#include <libdtdef.h>
#include <lnmdef.h>
#include <psldef.h>
#include <rmidef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* required (to compile) for VMS V7.3-1 and earlier */
#ifndef RMI$_CPUCOMPAT
#define RMI$_CPUCOMPAT 21193
#define RMI$_CPUINTSTK 21194
#define RMI$_CPUMPSYNCH 21195
#define RMI$_CPUKERNEL 21196
#define RMI$_CPUEXEC 21197
#define RMI$_CPUSUPER 21198
#define RMI$_CPUUSER 21199
#define RMI$_CPUIDLE 21200
#define RMI$_PROCBATCNT 21202
#define RMI$_PROCINTCNT 21203
#define RMI$_PROCNETCNT 21204
#define RMI$_MSCP_EVERYTHING 4265
#endif

/* required for VMS V7.3-1 and earlier */
#ifdef  __VAX
globalref sys$gw_ijobcnt;
globalref sys$gw_bjobcnt;
globalref sys$gw_njobcnt;
#else
globalref sys$gl_ijobcnt;
globalref sys$gl_bjobcnt;
globalref sys$gl_njobcnt;
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

#define BOOL int
#define TRUE 1
#define FALSE 0
 
#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))
 
#define FI_LI "MONDESI", __LINE__
#define EXIT_FI_LI(status) { printf ("[%s:%d]", FI_LI); exit(status); }

#define ALERT_SECONDS 90
#define AMBIT_PRIORITY 6
#define REFRESH_INTERVAL 2

#define BAR_HEIGHT 14
#define BAR_WIDTH (GRAPH_WIDTH+2 + GRAPH_WIDTH+2 + 10)

#define NODE_HEIGHT ((BAR_HEIGHT * 9) + 20)
#define NODE_HEIGHT_SCS ((BAR_HEIGHT * 12) + 20)
#define NODE_WIDTH 952

#define GRAPH_HEIGHT 100
#define GRAPH_WIDTH 300

#define HIST_HEIGHT ((GRAPH_HEIGHT * 3) + (10 * 3))
#define HIST_HEIGHT_SCS ((GRAPH_HEIGHT * 4) + (10 * 4))

/* exit after fifteen minutes without a WebSocket connection */
#define EXIT_AFTER_SECONDS 15 * 60;

/* maximum number of network devices */
#define NI_DEVICE_MAX 8

/*******************/
/* data structures */
/*******************/

#ifndef __VAX
#   pragma member_alignment __save
#   pragma nomember_alignment
#endif

typedef struct STRUCT_IOSBLK
{
   unsigned short  iosb$w_status;
   unsigned short  iosb$w_bcnt;
   unsigned int    iosb$l_reserved;
}
IOSBLK;

#ifndef __VAX
#   pragma member_alignment __restore
#endif

typedef struct STRUCT_SYSTEM_DATA
{
   unsigned long  ActiveCpuCount,
                  AvailCpuCount,
                  BufferedIO,
                  ClusterMsgSent,
                  ClusterMsgRcvd,
                  ClusterKbyteMap,
                  ClusterKbyteSent,
                  ClusterKbyteRcvd,
                  ClusterMember,
                  ClusterEVotes,
                  ClusterNodes,
                  ClusterQuorum,
                  ClusterVotes,
                  CpuCount,
                  DirectIO,
                  FcpRead,
                  FcpWrite,
                  FreeList,
                  HardPageFaults,
                  LockActivity,
                  LockWait,
                  MemInUse,
                  MemInUseMax,
                  MemSize,
                  ModList,
                  MscpRead,
                  MscpWrite,
                  PrevHardPageFaults,
                  PageFaults,
                  PageFaultsHard,
                  PageFaultsSoft,
                  PageFileFree,
                  PageFilePage,
                  PageFilePercent,
                  PageReadIO,
                  PageSize,
                  PageWriteIO,
                  ProcessCount,
                  ProcessBatCount,
                  ProcessIntCount,
                  ProcessNetCount,
                  Processes;
 
#ifndef __VAX
   unsigned __int64  NetIntCount,
                     NetIntBlocksRx,
                     NetIntBlocksTx,
                     NetIntBytesRx,
                     NetIntBytesTx,
                     NetIntErrorsHard,
                     NetIntErrorsSoft;
#else
   unsigned long  NetIntCount,
                  NetIntBlocksRx,
                  NetIntBlocksTx,
                  NetIntBytesRx,
                  NetIntBytesTx,
                  NetIntErrorsHard,
                  NetIntErrorsSoft;
#endif

   char  ArchName [15+1],
         HwName [60+1],
         NodeName [15+1],
         VmsVersion [8+1];

   unsigned long  BootBinTime [2],
                  CpuTicksBusy [2],
                  CpuTicksUser [2];

   int  CpuBusyPercent,
        CpuUserPercent;
}
SYSTEM_DATA;

/* used for multiple per-WebSocket client and CGI client (for convenience) */
typedef struct STRUCT_CLIENT_DATA
{
   BOOL  PrevInit,
         ReportCluster;

   int  InputCount,
        PrevBufferedIO,
        PrevBusyCpu,
        PrevClusterMsgSent,
        PrevClusterMsgRcvd,
        PrevClusterKbyteMap,
        PrevClusterKbyteSent,
        PrevClusterKbyteRcvd,
        PrevClusterNodes,
        PrevConnectedCount,
        PrevDirectIO,
        PrevFcpRead,
        PrevFcpWrite,
        PrevLockActivity,
        PrevLockWait,
        PrevMemInUse,
        PrevMinute,
        PrevMscpRead,
        PrevMscpWrite,
        PrevPageFaults,
        PrevPageFilePercent,
        PrevProcessCount,
        PrevProcessIntCount,
        PrevUserCpu,
        QueuedInput,
        QueuedOutput,
        RefreshSeconds,
        WebSocketScript;

#ifndef __VAX
   unsigned __int64  PrevNetIntBlocksRx,
                     PrevNetIntBlocksTx,
                     PrevNetIntBytesRx,
                     PrevNetIntBytesTx;
#else
   int  PrevNetIntBlocksRx,
        PrevNetIntBlocksTx,
        PrevNetIntBytesRx,
        PrevNetIntBytesTx;
#endif

   unsigned short  InputChannel,
                   OutputChannel;

   char  InputBuffer [1024],
         InputDevName [64],
         NakedBuffer [1024],
         OutputBuffer [2048],
         OutputDevName [64];

   struct STRUCT_IOSBLK  InputIOsb,
                         OutputIOsb;
} CLIENT_DATA;

/******************/
/* global storage */
/******************/

BOOL  Debug,
      ApacheHttpd,
      OsuHttpd;

int  ConnectedCount,
     IsCgiPlus,
     PrevPrio,
     UsageCount;

unsigned int  CollectDataSeconds,
              TimeStamp,
              ExitTimeStamp;

unsigned long  CurrentBinTime [2];
unsigned short  CurrentNumTime [7];

float  DeltaSeconds = 999999.0;

char  *CgiHttpRefererPtr,
      *CgiScriptNamePtr,
      *CgiQueryStringPtr,
      *CgiPlusEofPtr = NULL,
      *CgiPlusEotPtr = NULL,
      *CgiPlusEscPtr = NULL;

char  NetIntDev [NI_DEVICE_MAX*24],
      NotePad [128],
      RefererHost [128];

SYSTEM_DATA  SystemData;

/**************/
/* prototypes */
/**************/

void CollectDataPerSecond ();
int CollectNetInt (SYSTEM_DATA*);
int CollectRMI (SYSTEM_DATA*);
int CollectSyiStatic (SYSTEM_DATA*);
int CollectSyiDynamic (SYSTEM_DATA*);
char* GetCgiVar (char*);
char* GetCgiVarNull (char*);

void MonDeSiData (int);
void MonDeSiForm (char*, int, int);
void MonDeSiNode (char*, int);
void MonDeSiPage (char*, int, int);
int MonDeSiPrepareData (CLIENT_DATA *clptr);
void ReturnPrio ();

void WebSockRead (CLIENT_DATA*);
void WebSockReadAst (CLIENT_DATA*);
void WebSockRequest (int);
void WebSockRestoreLifetime ();
void WebSockOpen (CLIENT_DATA*);
void WebSockWrite (CLIENT_DATA*, char*, int);
void WebSockWriteAst (CLIENT_DATA*);
int WebSockUnFrame (char*);
void WebSockUpdate (CLIENT_DATA*);
void WebSockWait ();

/*****************************************************************************/
/*
*/

int main ()

{
   int  status,
        NodesAcross,
        RefreshInterval;
   char  *cptr, *sptr, *zptr,
         *NodeList;
   char  IframeId [64];

   /*********/
   /* begin */
   /*********/

   /* ambit attempt to bump the process priority (if INSTALLed with ALTPRI) */
   atexit (&ReturnPrio);
   sys$setpri (0, 0, AMBIT_PRIORITY, &PrevPrio, 0, 0);

   if (getenv ("MONDESI$DBUG")) Debug = 1;

   if (Debug)
      fprintf (stdout,
"Status: 200\n\
Content-Type: text/plain\n\
Script-Control: X-content-encoding-gzip=0\n\
\r\n");

   ApacheHttpd = (getenv ("APACHE$SHARED_SOCKET") != NULL ||
                  getenv ("APACHE$PARENT_PID") != NULL ||
                  getenv ("APACHE$SERVER_TAG") != NULL);
   OsuHttpd = (getenv("WWWEXEC_RUNDOWN_STRING") != NULL);

   if (!Debug && !OsuHttpd)
      if (!(stdout = freopen ("SYS$OUTPUT:", "w", stdout, "ctx=bin")))
         EXIT_FI_LI (vaxc$errno);                                           

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);
   if ((CgiPlusEscPtr = getenv("CGIPLUSESC")) == NULL) CgiPlusEscPtr = "";
   if ((CgiPlusEotPtr = getenv("CGIPLUSEOT")) == NULL) CgiPlusEotPtr = "";

   memset (&SystemData, 0, sizeof(SystemData));
   CollectSyiStatic (&SystemData);
   SystemData.NetIntCount = CollectNetInt (NULL);
   CollectDataPerSecond ();

   do {

      /* with CGIplus this call will block waiting for the next request */
      GetCgiVar ("");

      /********************/
      /* request received */
      /********************/

      UsageCount++;

      CgiHttpRefererPtr = GetCgiVar("HTTP_REFERER");
      CgiScriptNamePtr = GetCgiVar("SCRIPT_NAME");
      CgiQueryStringPtr = GetCgiVar("QUERY_STRING");

      /* derive the URL scheme and host from the HTTP referrer field */
      zptr = (sptr = RefererHost) + sizeof(RefererHost)-1;
      cptr = CgiHttpRefererPtr;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';

      if (isdigit(*CgiQueryStringPtr))
      {
         /********/
         /* data */
         /********/

         RefreshInterval = atoi(CgiQueryStringPtr);
         if (RefreshInterval <= 0 || RefreshInterval > 60)
            RefreshInterval = REFRESH_INTERVAL;

         if (GetCgiVarNull ("WEBSOCKET_INPUT"))
         {
            /* via WebSocket connected client */
            WebSockRequest (RefreshInterval);
         }
         else
         {
            /* via HTTP connected client */
            MonDeSiData (RefreshInterval);
         }
      }
      else
      {
         /********/
         /* page */
         /********/

         NodesAcross = 1;
         if (cptr = strstr (CgiQueryStringPtr, "across="))
            NodesAcross = atoi(cptr+7);
         if (NodesAcross < 0 || NodesAcross > 4) NodesAcross = 1;

         IframeId[0] = '\0';
         if (cptr = strstr (CgiQueryStringPtr, "iframe="))
         {
            zptr = (sptr = IframeId) + sizeof(IframeId)-1;
            for (cptr += 7;
                 *cptr && *cptr != '&' && sptr < zptr;
                 *sptr++ = *cptr++);
            *sptr = '\0';
         }

         if (cptr = strstr (CgiQueryStringPtr, "interval="))
            RefreshInterval = atoi(cptr+9);
         if (RefreshInterval <= 0 || RefreshInterval > 60)
            RefreshInterval = REFRESH_INTERVAL;
                      
         if (NodeList = strstr (CgiQueryStringPtr, "nodes="))
         {
            for (sptr = (NodeList += 6); *sptr && *sptr != '&'; sptr++);
            *sptr = '\0';
            UrlDecode (NodeList);
            /* eliminate those pesky W*s <CR><LF>s */
            for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
               if (*(unsigned short*)cptr == '\r\n') cptr++;
            *sptr = '\0';
            /* compress white-space */
            for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
               while (isspace(*cptr) && isspace(*(cptr+1))) cptr++;
            *sptr = '\0';
         }
         else
            NodeList = "";

         if (ispunct(*CgiQueryStringPtr))
            MonDeSiForm (NodeList, NodesAcross, RefreshInterval);
         else
         if (IframeId[0])
            MonDeSiNode (IframeId, RefreshInterval);
         else
         if (*NodeList)
            MonDeSiPage (NodeList, NodesAcross, RefreshInterval);
         else
            MonDeSiPage (SystemData.NodeName, 1, RefreshInterval);
      }

      if (IsCgiPlus)
      {
         /* send CGIplus end-of-request sentinal */
         fflush (stdout);
         fputs (CgiPlusEofPtr, stdout);
         fflush (stdout);
      }

   } while (IsCgiPlus);

   exit (SS$_NORMAL);
} 

/*****************************************************************************/
/*
Return to previous process priority (in case ambit raise was successful).
*/

void ReturnPrio ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (!PrevPrio) return;

   status = sys$setpri (0, 0, PrevPrio, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$setpri() %%X%08.08X\n", status);
} 

/*****************************************************************************/
/*
Provide the client with the configuration setup form.
*/

void MonDeSiForm
(
char *NodeList,
int NodesAcross,
int RefreshInterval
)
{
   char  *cptr, *sptr;
   char  NodeUrl [256];

   /*********/
   /* begin */
   /*********/

   if (!*NodeList)
   {
      /* concoct a node list entry */
      sptr = GetCgiVar ("REQUEST_SCHEME");
      if (!*sptr)
      {
         sptr = GetCgiVar ("SERVER_PORT");
         if (atoi(sptr) == 443)
            sptr = "https:";
         else
            sptr = "http:";
      }
      cptr = GetCgiVar ("HTTP_HOST");
      sprintf (NodeList = NodeUrl, "%s//%s%s", sptr, cptr, CgiScriptNamePtr);
   }

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
\
<style type=\"text/css\">\n\
\
body { background-color:DarkSlateGray; margin:1em; \
font-family: arial, helvetica, tahoma; text-align:center; font-size: 12pt; }\n\
input { border-width:1px; border-style:outset; }\n\
textarea,select { border-width:1px; border-style:inset; }\n\
.pageTitle { font-size:1.4em; font-weight:bold; letter-spacing:1px; \
text-decoration:underline; color:white; }\n\
.subTitle { font-size:0.9em; font-style:italic; letter-spacing:1px; \
color:white; }\n\
\
</style>\n\
\
<title>%s</title>\n\
</head>\n\
<body>\n\
<form method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
\
<span class=\"pageTitle\">Monitor de Sistema</span>\n\
<p style=\"margin-top:0.5em;\">\
<span class=\"subTitle\">~ System Monitor ~</span>\n\
\
<p><textarea id=\"nodes\" name=\"nodes\" rows=\"10\" cols=\"64\" \
scroll=\"auto\">\
%s</textarea>\n\
\
<p><select name=\"across\">\n\
<option value=\"1\"%s>1\n\
<option value=\"2\"%s>2\n\
<option value=\"3\"%s>3\n\
<option value=\"4\"%s>4\n\
</select>\
<span style=\"color:white;\">&nbsp;&nbsp;node(s) across window</span>\n\
\
<p><span style=\"color:white;\">Sample rate&nbsp;&nbsp;</span>\
<select name=\"interval\">\n\
<option value=\"0\"%s>off\n\
<option value=\"1\"%s>1\n\
<option value=\"2\"%s>2\n\
<option value=\"5\"%s>5\n\
<option value=\"15\"%s>15\n\
</select>\
<span style=\"color:white;\">&nbsp;&nbsp;seconds</span>\n\
\
<p><input type=\"submit\" value=\" Monitor \">&nbsp;&nbsp;\n\
<input type=\"reset\" value=\" Reset \" \
onclick=\"return confirm(\'Are you sure?\')\">&nbsp;&nbsp;\n\
<input type=\"reset\" value=\" Clear \" \
onclick=\"if (confirm(\'Are you sure?\')) \
document.getElementById(\'nodes\').value=\'\'; return false;\">\n\
\
</form>\n\
</body>\n\
</html>\n",
            SoftwareID,
            CgiScriptNamePtr,
            /* should be no HTML-forbidden characters in this! */
            NodeList,
            NodesAcross == 1 ? " selected" : "",
            NodesAcross == 2 ? " selected" : "",
            NodesAcross == 3 ? " selected" : "",
            NodesAcross == 4 ? " selected" : "",
            RefreshInterval ==  0 ? " selected" : "",
            RefreshInterval ==  1 ? " selected" : "",
            RefreshInterval ==  2 ? " selected" : "",
            RefreshInterval ==  5 ? " selected" : "",
            RefreshInterval == 15 ? " selected" : "");
} 

/*****************************************************************************/
/*
Provide the page that contains one or more monitor panels (IFRAMEs containing
their own HTML 'page' content).
*/

void MonDeSiPage
(
char *NodeList,
int NodesAcross,
int RefreshInterval
)
{
   int  NodeCount = 0;
   char  *aptr, *cptr, *sptr, *zptr;
   char  HostUrl [128];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiPage() |%s|\n", NodeList);

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
\
<style type=\"text/css\">\n\
\
body { background-color:DarkSlateGray; margin:1em; \
font-family: arial, helvetica, tahoma; text-align:center; font-size: 12pt; }\n\
@media print { body { background-color:white; } }\n\
input { border-width:1px; border-style:outset; }\n\
@media print { input { visibility:hidden; } }\n\
.node_iframe { height:%dpx; width:%dpx; border:0; padding:0; margin:5px; }\n\
.nodes_across { border:0; padding:0; margin:0; }\n\
.nodes_across td { vertical-align:top; }\n\
\
</style>\n\
\
<script language=\"JavaScript\">\n\
\
function mdsMsgChannel (msg)\n\
{\n\
   // message posting browsers (HTML5)\n\
   if (msg.data.substring(0,3) == \"mds\")\n\
   {\n\
      eval(msg.data);\n\
      return;\n\
   }\n\
   if (msg.data != \"\") alert(\"ERROR:\"+msg.data);\n\
}\n\
\
try { var mdsMessageListener = \
window.addEventListener(\"message\",mdsMsgChannel,false); } \
catch(err) { }\n\
\
function mdsBackChannel (forEval)\n\
{\n\
   // inter-frame kludge\n\
   if (forEval.substring(0,4) == \"#mds\")\n\
   {\n\
      eval(forEval.substr(1));\n\
      return;\n\
   }\n\
   if (forEval != \"\") alert(\"ERROR:\"+forEval.substr(1));\n\
}\n\
\
function mdsIframeHeight (iframeId,toTop,height)\n\
{\n\
   // remotely called via IFC\n\
   var oNodeFrame = document.getElementById(iframeId);\n\
   oNodeFrame.height = oNodeFrame.style.height = height + \"px\";\n\
   if (toTop) try { if (oNodeFrame.offsetTop > 100) \
window.scrollTo(0,oNodeFrame.offsetTop-10); } catch (err) { }\n\
}\n\
\
</script>\n\
<title>%s</title>\n\
</head>\n\
<body>\n",
            NODE_HEIGHT, NODE_WIDTH,
            SoftwareID);

   cptr = NodeList;
   while (*cptr)
   {
      while (*cptr && isspace(*cptr)) cptr++;
      zptr = (sptr = HostUrl) + sizeof(HostUrl)-1;
      /* parse to the end of the scheme */
      while (*cptr && !isspace(*cptr) && *cptr != '/' && sptr < zptr)
         *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /* if no domain name */
      if (!strchr (HostUrl, '.'))
      {
         /* same host, just use the script component */
         sptr = HostUrl;
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      else
      /* if provides a script component */
      if (*cptr && !isspace(*cptr))
      {
         if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
         while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      }
      else
      {
         /* append a script component */
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      *sptr = '\0';
      while (*cptr && !isspace(*cptr)) cptr++;
      while (*cptr && isspace(*cptr)) cptr++;

      if (NodesAcross > 1)
      {
         if (!(NodeCount % NodesAcross))
         {
            if (!NodeCount)
               fputs ("<table class=\"nodes_across\">\n<tr><td>\n", stdout);
            else
               fputs ("</td></tr>\n<tr><td>\n", stdout);
         }
         else
            fputs ("</td><td>\n", stdout);
      }
      NodeCount++,

      fprintf (stdout,
"<iframe id=\"node%d\" class=\"node_iframe\" \
frameborder=\"0\" scrolling=\"no\" \
src=\"%s?iframe=node%d&interval=%d\"></iframe>\n",
               NodeCount,
               HostUrl,
               NodeCount,
               RefreshInterval);
   }

   if (NodesAcross > 1) fputs ("</td></tr>\n</table>\n", stdout);

   fprintf (stdout,
"<form method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
<input type=\"hidden\" name=\"!\" value=\"\">\n\
<input type=\"hidden\" name=\"nodes\" value=\"%s\">\n\
<input type=\"hidden\" name=\"across\" value=\"%d\">\n\
<input type=\"hidden\" name=\"interval\" value=\"%d\">\n\
<input type=\"submit\" value=\" Configure \">\n\
</form>\n\
</body>\n\
</html>\n",
            CgiScriptNamePtr,
            strchr(NodeList,'.') ? NodeList : "",
            NodesAcross,
            RefreshInterval);
} 

/*****************************************************************************/
/*
Provide the in-frame node display.
*/

void MonDeSiNode
(
char *IframeId,
int RefreshInterval
)
{
   int  WebSocketOk = 0;
   char  *cptr;
   char  DivStyle [256];

   /*********/
   /* begin */
   /*********/

   /* if WASD and the version >= 10.1 then WebSocket can be used */
   if (cptr = GetCgiVarNull("SERVER_SOFTWARE"))
      if (cptr = strstr (cptr, "WASD/10."))
         if (WebSocketOk = (atoi(cptr+8) >= 1))
            if (cptr = getenv ("MONDESI_WEBSOCKET"))
               WebSocketOk = atoi(cptr);

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
\
<style type=\"text/css\">\n\
\
body { background-color:white; margin:0; padding:0; \
font-family: arial, helvetica, tahoma; font-size:12pt; }\n\
.node_block { width:100%%; background-color:beige; color:black; \
border-style:outset; border-width:1px; border-color:black; \
margin:0; padding:0; text-align:left; }\n\
@media print { .node_block { background-color:white; \
border-style:solid; border-width:2px; } }\n\
.node_block td { margin:10px; padding:0; }\n\
.node_graph { border-width:0; margin:0; padding:0; }\n\
.node_graph td { margin:0; padding:0; }\n\
.node_notes { vertical-align:top; }\n\
.node_name { font-size:1.2em; font-weight:bold; text-decoration:underline; }\n\
.node_text { font-size:0.8em; }\n\
.node_control input,select { font-size:0.7em; border-style:outset; \
border-width:1px; border-color:gray; width:4.2em; \
padding:0; margin-right:0.4em; }\n\
@media print { .node_notes input { visibility:hidden; } }\n\
.history_graph { border-width:0; margin:0; padding:0 0 5px 0; }\n\
.history_graph td { margin:0; padding:0; }\n\
\
</style>\n\
<script language=\"JavaScript\">\n\
\
var WASDwebSocket = %s;\n\
var mdsAlertSecs = %d;\n\
var mdsRefreshInterval = %d;\n\
var mdsRefererHost = \"%s\"\n\
var mdsScriptName = \"%s\";\n\
var mdsNodeName = \"%s\";\n\
var mdsCpuCount = %d;\n\
var mdsCluster = %s;\n\
var mdsIframeId = \"%s\";\n\
var mdsBarWidth = %d;\n\
var mdsBarHeight = %d;\n\
var mdsGraphWidth = %d;\n\
var mdsGraphHeight = %d;\n\
var mdsNodeHeight = %d;\n\
var mdsNodeHeightScs = %d;\n\
var mdsHistHeight = %d;\n\
var mdsHistHeightScs = %d;\n\
var mdsHistVisible = false;\n\
var mdsBody;\n\
\
</script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
\
<title>%s</title>\n\
</head>\n\
<body>\n",
            WebSocketOk ? "true" : "false",
            ALERT_SECONDS,
            RefreshInterval,
            RefererHost,
            CgiScriptNamePtr,
            SystemData.NodeName,
            SystemData.ActiveCpuCount,
            SystemData.ClusterMember ? "true" : "false",
            IframeId,
            BAR_WIDTH, BAR_HEIGHT,
            GRAPH_WIDTH, GRAPH_HEIGHT,
            NODE_HEIGHT, NODE_HEIGHT_SCS,
            HIST_HEIGHT, HIST_HEIGHT_SCS,
            RefererHost, "/mondesi/-/js/mondesi.js",
            RefererHost, "/mondesi/-/js/wz_jsgraphics.js",
            SoftwareID);

   fprintf (stdout, "<table class=\"node_block\">\n<tr><td>\n");

   /**************/
   /* bar graphs */
   /**************/

   sprintf (DivStyle,
"position:relative;height:%dpx;width:%dpx;margin-left:5px;",
            BAR_HEIGHT, BAR_WIDTH);

   fprintf (stdout,
"<table id=\"node_graph\" class=\"node_graph\">\n\
<tr><td height=\"100%%\"><div id=\"node_notes\" \
style=\"width:%dpx;margin:5px 5px 0 5px;\"></div></td>\n\
\
<td class=\"node_notes\" rowspan=\"2\">\n\
<div id=\"node_cpu\" style=\"%smargin-top:5px;\"></div>\n\
<div id=\"node_mem\" style=\"%s\"></div>\n\
<div id=\"node_pgfl\" style=\"%sdisplay:none;\"></div>\n\
<div id=\"node_flts\" style=\"%s\"></div>\n\
<div id=\"node_bio\" style=\"%s\"></div>\n\
<div id=\"node_dio\" style=\"%s\"></div>\n\
<div id=\"node_lock\" style=\"%s\"></div>\n\
<div id=\"node_ni\" style=\"%s\"></div><!-- %s -->\n\
<div id=\"node_disk\" style=\"%s\"></div>\n\
<div id=\"node_prc\" style=\"%s\"></div>\n",
            GRAPH_WIDTH+2,
            DivStyle, DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            DivStyle, NetIntDev, DivStyle, DivStyle);

   if (SystemData.ClusterMember)
      fprintf (stdout,
"<div id=\"node_scs\" style=\"display:none;\">\n\
<div id=\"node_scsmg\" style=\"%s\"></div>\n\
<div id=\"node_scskb\" style=\"%s\"></div>\n\
<div id=\"node_mscp\" style=\"%s\"></div>\n\
</div>\n",
               DivStyle, DivStyle, DivStyle);

   fprintf (stdout,
"<div id=\"node_debug\"></div>\n\
</td></tr>\n\
\
<tr><td style=\"height:100%%;vertical-align:bottom;padding-bottom:0.5em;\">\
<div id=\"node_control\" class=\"node_control\" \
style=\"width:%dpx;height:1em;margin:0 0 0 5px;white-space:nowrap;\">\n",
            GRAPH_WIDTH+2);

   fprintf (stdout,
"<input type=\"hidden\" name=\"refresh\" value=\"%d\">\n\
<input id=\"history_btn\" type=\"submit\" value=\"history\" \
onclick=\"mdsHistoryClick()\" style=\"width:4.5em;\">\
<input id=\"scs_btn\" type=\"submit\" value=\"cluster\" \
onclick=\"mdsClusterClick()\" style=\"width:4.5em;\" disabled>\
<input id=\"print_btn\" type=\"submit\" value=\"print\" \
onclick=\"mdsPrintNode();return false;\" style=\"width:4.0em;\">\
<input type=\"submit\" value=\"scale\" \
onclick=\"return mdsScale();\" style=\"width:4.0em;\">\
<input type=\"submit\" value=\"zero\" \
onclick=\"return mdsNodeZero();\" style=\"width:3.5em;\">\
<input id=\"start_stop_btn\" type=\"submit\" value=\"stop\" \
onclick=\"return mdsStartStop();\" style=\"width:3.5em;\">\n",
            RefreshInterval);

   fprintf (stdout,
"</div></td></tr>\n\
</table>\n\
</td></tr>\n\
<tr><td style=\"padding-top:5px;\">\n");

   /******************/
   /* history graphs */
   /******************/

   sprintf (DivStyle,
"position:relative;height:%dpx;width:%dpx;margin:0 5px 5px 5px;",
            GRAPH_HEIGHT+2, GRAPH_WIDTH+2);

   fprintf (stdout,
"<table id=\"history_graph\" class=\"history_graph\">\n\
\
<tr><td>\n\
<div id=\"hist_cpu\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_mem\" style=\"%s\" \
onmouseover=\"mdsMemPgfl(true)\"></div>\n\
<div id=\"hist_pgfl\" style=\"%sdisplay:none;\" \
onmouseout=\"mdsMemPgfl(false)\"></div>\n\
</td><td>\n\
<div id=\"hist_flts\" style=\"%s\"></div>\n\
</td></tr>\n\
\
<tr><td>\n\
<div id=\"hist_bio\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_dio\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_lock\" style=\"%s\"></div>\n\
</td></tr>\n\
\
<tr><td>\n\
<div id=\"hist_ni\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_disk\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_prc\" style=\"%s\"></div>\n\
</td></tr>\n",
            DivStyle, DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle);

   if (SystemData.ClusterMember)
      fprintf (stdout,
"<tr id=\"hist_scs\" style=\"display:none;\"><td>\n\
<div id=\"hist_scsmg\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_scskb\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_mscp\" style=\"%s\"></div>\n\
</td></tr>\n",
               DivStyle, DivStyle, DivStyle);

   fputs ("</table>\n</td></tr>\n</table>\n", stdout);

   /**************/
   /* javascript */
   /**************/

   fprintf (stdout,
"<script language=\"JavaScript\">\n\
\
mdsBody = document.getElementsByTagName(\"body\").item(0);\n\
mdsInitialStart();\n\
mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\");\n\
\
mdsNodeInit(\"cpu\",true);\n\
mdsNodeInit(\"mem\",true);\n\
mdsNodeInit(\"pgfl\",true);\n\
mdsNodeInit(\"flts\",true);\n\
mdsNodeInit(\"bio\",true);\n\
mdsNodeInit(\"dio\",true);\n\
mdsNodeInit(\"lock\",true);\n\
mdsNodeInit(\"ni\",%s);\n\
mdsNodeInit(\"disk\",true);\n\
mdsNodeInit(\"prc\",true);\n",
            SystemData.NodeName,
            SystemData.NetIntCount ? "true" : "false");

   if (SystemData.ClusterMember)
      fprintf (stdout,
"mdsNodeInit(\"scsmg\",true);\n\
mdsNodeInit(\"scskb\",true);\n\
mdsNodeInit(\"mscp\",true);\n");

   fputs ("</script>\n</body>\n</html>\n", stdout);
} 

/*****************************************************************************/
/*
For an HTTP request.  At the defined interval gather dynamic system and
performance data and for appropriate (changed) data call the relevant
JavaScript (RPC) function in the parent panel (window) to process and graph
associated data.  Do the best we can to disable proxy caching ensuring a timely
flow of response data.
*/

void MonDeSiData (int RefreshInterval)

{
   int  status,
        HostUrlLength,
        TimeStamp;
   char  *cptr, *sptr, *zptr;
   CLIENT_DATA  ClientData;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiData()\n");

   /* derive the host URL from the HTTP referrer field */
   cptr = CgiHttpRefererPtr;
   while (*cptr && *(unsigned short*)cptr != '//') cptr++;
   if (*cptr) cptr += 2;
   while (*cptr && *cptr != '/') cptr++;
   HostUrlLength = cptr - CgiHttpRefererPtr;

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Script-Control: X-content-encoding-gzip=0\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<body onload=\"parent.mdsInitialStart();\">\n");
   fflush (stdout);

   memset (&ClientData, 0, sizeof(ClientData));

   /* infinite loop */
   for(;;)
   {
      if (MonDeSiPrepareData (&ClientData))
      {
         fputs (ClientData.OutputBuffer, stdout);
         fflush (stdout);
      }
      sleep (RefreshInterval);
   }
   /* should never ... */
} 

/*****************************************************************************/
/*
Ued for both HTTP and WebSocket requests.
Writes the data update JavaScript into the client buffer.
Returns the number of characters in that buffer.
*/

int MonDeSiPrepareData (CLIENT_DATA *clptr)

{
   static $DESCRIPTOR (DateTimeFaoDsc, "!17%D\0");
   static $DESCRIPTOR (UpTimeFaoDsc, "!10%D\0");

   int  status,
        HostUrlLength;
   unsigned long  UpBinTime [2];
   char  *cptr, *sptr, *zptr;
   char  CpuActive [32],
         DateTime [32],
         MemSize [32],
         Stupid_njobcnt [16],
         UpTime [32],
         WebSocketStatus [128];
   $DESCRIPTOR (DateTimeDsc, DateTime);
   $DESCRIPTOR (UpTimeDsc, UpTime);
   SYSTEM_DATA *sysptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiPrepareData()\n");

   if (!clptr->PrevInit)
   {
      /* initialise the previous value storage */
      clptr->PrevBufferedIO = -1;
      clptr->PrevBusyCpu = -1;
      clptr->PrevClusterMsgSent = -1;
      clptr->PrevClusterMsgRcvd = -1;
      clptr->PrevClusterKbyteMap = -1;
      clptr->PrevClusterKbyteSent = -1;
      clptr->PrevClusterKbyteRcvd = -1;
      clptr->PrevClusterNodes = 0;
      clptr->PrevDirectIO = -1;
      clptr->PrevNetIntBlocksRx = -1;
      clptr->PrevNetIntBlocksTx = -1;
      clptr->PrevNetIntBytesRx = -1;
      clptr->PrevNetIntBytesTx = -1;
      clptr->PrevFcpRead = -1;
      clptr->PrevFcpWrite = -1;
      clptr->PrevLockActivity = -1;
      clptr->PrevLockWait = -1;
      clptr->PrevMemInUse = -1;
      clptr->PrevMinute = -1;
      clptr->PrevMscpRead = -1;
      clptr->PrevMscpWrite = -1;
      clptr->PrevPageFaults = -1;
      clptr->PrevPageFilePercent = -1;
      clptr->PrevProcessCount = -1;
      clptr->PrevProcessIntCount = -1;
      clptr->PrevUserCpu = -1;

      /* plus decide on a strategy */
      clptr->WebSocketScript = (GetCgiVarNull("WEBSOCKET_INPUT") != NULL);
      if (!clptr->WebSocketScript) clptr->ReportCluster = 1;
   }

   zptr = (sptr = clptr->OutputBuffer) + sizeof(clptr->OutputBuffer)-1;

   if (clptr->WebSocketScript)
      *sptr++ = 0x00;
   else
      sptr += sprintf (sptr, "<script language=\"JavaScript\">");

   if (!clptr->PrevInit)
   {
      clptr->PrevInit = TRUE;
      /* only needed the once */
      sptr += sprintf (sptr, "the=%s;",
                       clptr->WebSocketScript ? "window" : "parent");
   }

   sysptr = &SystemData;

   if (clptr->PrevMinute != CurrentNumTime[4] ||
       sysptr->ProcessCount != clptr->PrevProcessCount ||
       sysptr->ClusterNodes != clptr->PrevClusterNodes ||
       clptr->PrevConnectedCount != ConnectedCount ||
       NotePad[0])
   {
      clptr->PrevMinute = CurrentNumTime[4];
      clptr->PrevConnectedCount = ConnectedCount;

      sys$fao (&DateTimeFaoDsc, 0, &DateTimeDsc, &CurrentBinTime);
      lib$sub_times (&CurrentBinTime, &sysptr->BootBinTime, &UpBinTime);
      sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &UpBinTime);

      if (sysptr->ActiveCpuCount == sysptr->AvailCpuCount)
         sprintf (CpuActive, "%d", sysptr->ActiveCpuCount);
      else
         sprintf (CpuActive, "%d/%d", sysptr->ActiveCpuCount,
                                      sysptr->AvailCpuCount);

      if (sysptr->MemSize > 1023)
         sprintf (MemSize, "%.0f&nbsp;GB", (float)sysptr->MemSize / 1024.0);
      else
         sprintf (MemSize, "%d&nbsp;MB", sysptr->MemSize);

      sptr += sprintf (sptr,
"mdsCpuCount=%d;\
the.mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\
<span class=\'node_text\'>\
<br>%s<br>%s %s&nbsp;CPU%s %s %s&nbsp;%s&nbsp <nobr>Up %s</nobr>",
                       sysptr->ActiveCpuCount,
                       sysptr->NodeName,
                       DateTime,
                       sysptr->HwName,
                       CpuActive,
                       sysptr->AvailCpuCount > 1 ? "s" : "",
                       MemSize,
                       sysptr->ArchName,
                       sysptr->VmsVersion,
                       UpTime);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (sysptr->ClusterMember)
      {
         sptr += sprintf (sptr,
"&nbsp Cluster nodes:%d votes:%d expect:%d quorum:%d",
                          sysptr->ClusterNodes,
                          sysptr->ClusterVotes,
                          sysptr->ClusterEVotes,
                          sysptr->ClusterQuorum);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         clptr->PrevClusterNodes = sysptr->ClusterNodes;
      }

      if ((int)sysptr->ProcessNetCount >= 0)
         sprintf (Stupid_njobcnt, "%d", sysptr->ProcessNetCount);
      else
         strcpy (Stupid_njobcnt, "?");

      sptr += sprintf (sptr,
"&nbsp; Processes:%d interactive:%d batch:%d network:%s",
                       sysptr->ProcessCount,
                       sysptr->ProcessIntCount,
                       sysptr->ProcessBatCount,
                       Stupid_njobcnt);
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if ((int)sysptr->PageFilePercent >= 50)
      {
         sptr += sprintf (sptr, "&nbsp; <span%s>PGFL:%d%%</span>",
                          (int)sysptr->PageFilePercent >= 85 ?
                             " style=\'font-weight:bold;color:red;\'" : "",
                          sysptr->PageFilePercent);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if (sysptr->NetIntErrorsSoft || sysptr->NetIntErrorsHard)
      {
         sptr += sprintf (sptr, "&nbsp; <span%s>NI:%d/%d</span>",
                          sysptr->NetIntErrorsHard ?
                             " style=\'font-weight:bold;color:red;\'" : "",
                          (int)sysptr->NetIntErrorsSoft,
                          (int)sysptr->NetIntErrorsHard);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
      }

      if (clptr->WebSocketScript)
         sprintf (WebSocketStatus,
                  "/<span style=\'font-size:0.8em\'>WebSocket:%d</span>",
                  ConnectedCount);
      else
         WebSocketStatus[0] = '\0';

      sptr += sprintf (sptr,
"&nbsp; <span style=\'color:gray\'>(v%s/%s%s%s%s)</span>",
                       SOFTWAREVN,
                       !ApacheHttpd && !OsuHttpd ? "WASD" : "",
                       WebSocketStatus,
                       ApacheHttpd ? "Apache" : "",
                       OsuHttpd ? "OSU" : "");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      if (NotePad[0])
      {
         sptr += sprintf (sptr, "&nbsp; <span style=\'color:red;\'>%s</span>",
                          NotePad[0] == '$' ? NotePad+1 : NotePad);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
         if (NotePad[0] != '$') NotePad[0] = '\0';
      }

      sptr += sprintf (sptr, "</span>\");");
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

      /* ensure memory in use and page file is refreshed */
      clptr->PrevMemInUse = clptr->PrevPageFilePercent = -1;
   }

   /* after enough samples to make it representative */
   if (CollectDataSeconds > 2)
   {
      if (sysptr->CpuBusyPercent != clptr->PrevBusyCpu ||
          sysptr->CpuUserPercent != clptr->PrevUserCpu)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"cpu\",%u,%u,%u,%u);",
                          TimeStamp,
                          sysptr->ActiveCpuCount*100,
                          sysptr->CpuBusyPercent,
                          sysptr->CpuBusyPercent - sysptr->CpuUserPercent);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevBusyCpu = sysptr->CpuBusyPercent;
         clptr->PrevUserCpu = sysptr->CpuUserPercent;
      }

      if (sysptr->MemInUse != clptr->PrevMemInUse)
      {
         sptr += sprintf (sptr,
#ifndef __VAX
                          "the.mdsMonVal(\"mem\",%u,%llu,%llu);",
                          TimeStamp,
                          (__int64)(sysptr->MemSize) << 20,
                          (__int64)(sysptr->MemInUse) << 20);
#else
                          "the.mdsMonVal(\"mem\",%u,%u,%u);",
                          TimeStamp,
                          sysptr->MemSize,
                          sysptr->MemInUse);
#endif
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevMemInUse = sysptr->MemInUse;
      }

      if (sysptr->PageFilePercent != clptr->PrevPageFilePercent)
      {
         sptr += sprintf (sptr,
#ifndef __VAX
                          "the.mdsMonVal(\"pgfl\",%u,%llu,%llu);",
                          TimeStamp,
                          (__int64)sysptr->PageFilePage *
                          (__int64)sysptr->PageSize,
                          ((__int64)sysptr->PageFilePage *
                           (__int64)sysptr->PageSize) -
                          ((__int64)sysptr->PageFileFree *
                           (__int64)sysptr->PageSize));
#else
                          "the.mdsMonVal(\"pgfl\",%u,%u,%u);",
                          TimeStamp,
                          sysptr->PageFilePage,
                          sysptr->PageFilePage - sysptr->PageFileFree);
#endif
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevPageFilePercent = sysptr->PageFilePercent;
      }

      if (sysptr->PageFaults != clptr->PrevPageFaults)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"flts\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->PageFaultsSoft + sysptr->PageFaultsHard,
                          sysptr->PageFaultsHard);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevPageFaults = sysptr->PageFaults;
      }

      if (sysptr->BufferedIO != clptr->PrevBufferedIO)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"bio\",%u,-1,%u);",
                          TimeStamp,
                          sysptr->BufferedIO);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevBufferedIO = sysptr->BufferedIO;
      }

      if (sysptr->DirectIO != clptr->PrevDirectIO)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"dio\",%u,-1,%u);",
                          TimeStamp,
                          sysptr->DirectIO);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevDirectIO = sysptr->DirectIO;
      }

      if (sysptr->LockActivity != clptr->PrevLockActivity ||
          sysptr->LockWait != clptr->PrevLockWait)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"lock\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->LockActivity,
                          sysptr->LockWait);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevLockActivity = sysptr->LockActivity;
         clptr->PrevLockWait = sysptr->LockWait;
      }

      if (sysptr->NetIntCount &&
          (sysptr->NetIntBytesRx != clptr->PrevNetIntBytesRx ||
           sysptr->NetIntBytesTx != clptr->PrevNetIntBytesTx))
      {
         sptr += sprintf (sptr,
#ifndef __VAX
                          "the.mdsMonVal(\"ni\",%u,-1,%llu,%llu);",
#else
                          "the.mdsMonVal(\"ni\",%u,-1,%u,%u);",
#endif
                          TimeStamp,
                          sysptr->NetIntBytesRx + sysptr->NetIntBytesTx,
                          sysptr->NetIntBytesRx);

         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevNetIntBytesRx = sysptr->NetIntBytesRx;
         clptr->PrevNetIntBytesTx = sysptr->NetIntBytesTx;
      }

      if (sysptr->FcpRead != clptr->PrevFcpRead ||
          sysptr->FcpWrite != clptr->PrevFcpWrite)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"disk\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->FcpRead + sysptr->FcpWrite,
                          sysptr->FcpWrite);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevFcpRead = sysptr->FcpRead;
         clptr->PrevFcpWrite = sysptr->FcpWrite;
      }

      if (sysptr->ProcessCount != clptr->PrevProcessCount ||
          sysptr->ProcessIntCount != clptr->PrevProcessIntCount)
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"prc\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->ProcessCount,
                          sysptr->ProcessIntCount);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevProcessCount = sysptr->ProcessCount;
         clptr->PrevProcessIntCount = sysptr->ProcessIntCount;
      }

      if (sysptr->ClusterMember &&
          clptr->ReportCluster &&
          (sysptr->ClusterMsgRcvd != clptr->PrevClusterMsgRcvd ||
           sysptr->ClusterMsgSent != clptr->PrevClusterMsgSent))
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"scsmg\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->ClusterMsgRcvd + sysptr->ClusterMsgSent,
                          sysptr->ClusterMsgSent);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevClusterMsgRcvd = sysptr->ClusterMsgRcvd;
         clptr->PrevClusterMsgSent = sysptr->ClusterMsgSent;
      }

      if (sysptr->ClusterMember &&
          clptr->ReportCluster &&
          (sysptr->ClusterKbyteMap != clptr->PrevClusterKbyteMap))
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"scskb\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->ClusterKbyteMap,
                          0);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevClusterKbyteMap = sysptr->ClusterKbyteMap;
      }

      if (sysptr->ClusterMember &&
          clptr->ReportCluster &&
          (sysptr->MscpRead != clptr->PrevMscpRead ||
           sysptr->MscpWrite != clptr->PrevMscpWrite))
      {
         sptr += sprintf (sptr,
                          "the.mdsMonVal(\"mscp\",%u,-1,%u,%u);",
                          TimeStamp,
                          sysptr->MscpRead + sysptr->MscpWrite,
                          sysptr->MscpWrite);
         if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);

         clptr->PrevMscpRead = sysptr->MscpRead;
         clptr->PrevMscpWrite = sysptr->MscpWrite;
      }
   }


   if (clptr->WebSocketScript)
   {
      if (sptr - clptr->OutputBuffer > 13)
         *sptr++ = 0xff;
      else
         sptr = clptr->OutputBuffer; 
   }
   else
   {
      if (sptr - clptr->OutputBuffer > 45)
         sptr += sprintf (sptr, "</script>\n");
      else
         sptr = clptr->OutputBuffer; 
      if (sptr > zptr) EXIT_FI_LI (SS$_BUGCHECK);
   }
   *sptr = '\0';

   if (clptr->PrevMinute != CurrentNumTime[4])
      clptr->PrevMinute = CurrentNumTime[4];

  return (sptr - clptr->OutputBuffer);
}

/*****************************************************************************/
/*
Indepdendent, timer AST driven update of the system performance data.
This will support multiple clients more efficiently.
The updated data are then read by other indepdendent client threads.
*/

void CollectDataPerSecond ()

{
   static unsigned long  OneSecondDelta [2] = { -10000000, -1 };
   static const long cvtf_mode = LIB$K_DELTA_SECONDS_F;

   static unsigned long  PrevBinTime [2];

   int  status;
   unsigned long  DiffBinTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectDataPerSecond()\n");

   sys$gettim (&CurrentBinTime);
   sys$numtim (&CurrentNumTime, &CurrentBinTime);
   TimeStamp = decc$fix_time (&CurrentBinTime);

   if (ExitTimeStamp && TimeStamp > ExitTimeStamp) exit (SS$_NORMAL); 

   /* calculate a floating point delta time */
   if (!PrevBinTime[0] && !PrevBinTime[1])
   {
      PrevBinTime[0] = CurrentBinTime[0];
      PrevBinTime[1] = CurrentBinTime[1];
   }
   lib$sub_times (&CurrentBinTime, &PrevBinTime, &DiffBinTime);
   PrevBinTime[0] = CurrentBinTime[0];
   PrevBinTime[1] = CurrentBinTime[1];
#ifdef __ia64
   lib$cvts_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#else
   lib$cvtf_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#endif
   if (Debug) fprintf (stdout, "DeltaSeconds: %f\n", DeltaSeconds);

   CollectSyiDynamic (&SystemData);
   CollectRMI (&SystemData);
   CollectNetInt (&SystemData);

   status = sys$setimr (0, &OneSecondDelta, CollectDataPerSecond, NULL, 0);

   if (VMSnok(status)) EXIT_FI_LI (status);

   CollectDataSeconds++;
}

/*****************************************************************************/
/*
Per-boot system data.
*/ 

int CollectSyiStatic (SYSTEM_DATA *sysptr)

{
   static unsigned long  ActiveCpuCnt,
                         AvailCpuCount,
                         MemSize,
                         PageSize;
   static unsigned long  BootBinTime [2];
   static unsigned short  HwNameLength;
   static unsigned char  ClusterMember;
   static char  ArchName [15+1],
                HwName [1+60],
                NodeName [15+1],
                VmsVersion [8+1];

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList[] =
   {
       { sizeof(AvailCpuCount), SYI$_AVAILCPU_CNT, &AvailCpuCount, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(ArchName), SYI$_ARCH_NAME, &ArchName, 0 },
       { sizeof(BootBinTime), SYI$_BOOTTIME, &BootBinTime, 0 },
       { sizeof(ClusterMember), SYI$_CLUSTER_MEMBER, &ClusterMember, 0 },
       { sizeof(HwName), SYI$_HW_NAME, &HwName, &HwNameLength },
       { sizeof(MemSize), SYI$_MEMSIZE, &MemSize, 0 },
       { sizeof(NodeName), SYI$_SCSNODE, &NodeName, 0 },
       { sizeof(PageSize), SYI$_PAGE_SIZE, &PageSize, 0 },
       { sizeof(VmsVersion), SYI$_VERSION, &VmsVersion, 0 },
       {0,0,0,0}
   };

   int  status;
   char  *cptr, *sptr, *zptr;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiStatic()\n");

   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &ItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSnok (status)) return (status);

   zptr = (sptr = sysptr->ArchName) + sizeof(sysptr->ArchName)-1;
   for (cptr = ArchName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   HwName[HwNameLength] = '\0';
   zptr = (sptr = sysptr->HwName) + sizeof(sysptr->HwName)-1;
   for (cptr = HwName; *cptr && sptr < zptr; *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = sysptr->NodeName) + sizeof(sysptr->NodeName)-1;
   for (cptr = NodeName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = sysptr->VmsVersion) + sizeof(sysptr->VmsVersion)-1;
   for (cptr = VmsVersion;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   if (!(sysptr->ClusterMember = ClusterMember))
      if (cptr = getenv("MONDESI_CLUSTER_MEMBER"))
         sysptr->ClusterMember = atoi(cptr);

   /* convert to MBytes */
   sysptr->MemSize = MemSize >> 7;
   sysptr->PageSize = PageSize;

   sysptr->BootBinTime[0] = BootBinTime[0];
   sysptr->BootBinTime[1] = BootBinTime[1];

   sysptr->AvailCpuCount = AvailCpuCount;
   sysptr->ActiveCpuCount = ActiveCpuCnt;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s| %d %d\n",
               sysptr->ArchName, sysptr->HwName,
               sysptr->NodeName, sysptr->VmsVersion,
               sysptr->AvailCpuCount, ClusterMember);

   return (status);
} 

/*****************************************************************************/
/*
Dynamic system data.
*/ 

int CollectSyiDynamic (SYSTEM_DATA *sysptr)

{
   static unsigned long  ActiveCpuCnt,
                         PageFileFree,
                         PageFilePage;
   static unsigned short  ClusterEVotes,
                          ClusterNodes,
                          ClusterQuorum,
                          ClusterVotes;

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ClusterItemList[] =
   {
       { sizeof(ClusterNodes), SYI$_CLUSTER_NODES, &ClusterNodes, 0 },
       { sizeof(ClusterQuorum), SYI$_CLUSTER_QUORUM, &ClusterQuorum, 0 },
       { sizeof(ClusterVotes), SYI$_CLUSTER_VOTES, &ClusterVotes, 0 },
       { sizeof(ClusterEVotes), SYI$_CLUSTER_EVOTES, &ClusterEVotes, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   }, NonItemList[] =
   {
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   };

   int  status;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiDynamic()\n");

   /* collect System Information */
   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            sysptr->ClusterMember ?
               (void*)&ClusterItemList :
               (void*)&NonItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
      return (status);
   }

   if (sysptr->ClusterMember)
   {
      sysptr->ClusterEVotes = ClusterEVotes;
      sysptr->ClusterNodes = ClusterNodes;
      sysptr->ClusterQuorum = ClusterQuorum;
      sysptr->ClusterVotes = ClusterVotes;
   }

   sysptr->ActiveCpuCount = ActiveCpuCnt;

   sysptr->PageFilePercent = 100 - (PageFileFree * 100 / PageFilePage);
   sysptr->PageFileFree = PageFileFree;
   sysptr->PageFilePage = PageFilePage;

   return (status);
} 

/*****************************************************************************/
/*
Collect System Performance Information.

Uses the system service $GETRMI() available for VMS V7.3 and later.

Cannot use RMI$_CPU* or RMI$_PROC*CNT on VMS V7.3-1 (and V7.3?) because they
don't seem to exist!  Fall back to handcrafting CPU from RMI$_MODES and
sys$gl_*jobcnt (many thanks to Graham Burley for the suggestion).  This
approach maintains two item lists and if the V7.3-2 (and later) fails with a
SS$$_BADPARAM switches to the other list.  Subsequent alternate processing is
performed by testing which list is in use.
*/ 

int CollectRMI (SYSTEM_DATA *sysptr)

{
/* required for VMS V7.3-1 and earlier */
#define INTSTK_MODE  0
#define MPSYNCH_MODE 1
#define KERNEL_MODE  2
#define EXEC_MODE    3
#define SUPER_MODE   4
#define USER_MODE    5
/* 6 is reserved */
#define IDLE_MODE    7

#define MODE_COUNT IDLE_MODE+1
#define CPU_MAX             32

#pragma member_alignment save 
#pragma nomember_alignment 

   /* structure for a single CPU's mode ticks */
   static struct SingleCPU {
       unsigned char  CPUid;
       unsigned long  Count [MODE_COUNT];
   };

   /* a structure all CPU's mode ticks for a system with up to 32 CPUs */
   static struct {
       unsigned long  NumberOfCPUs;
       struct SingleCPU  CPU [CPU_MAX];
   } ModeTicks;

#pragma member_alignment restore

   static unsigned long  BufferedIO,
                         ClusterMsgSent,
                         ClusterMsgRcvd,
                         ClusterKbyteMap,
                         ClusterKbyteSent,
                         ClusterKbyteRcvd,
                         DirectIO,
                         FcpRead,
                         FcpWrite,
                         LockConvert,
                         LockDequeue,
                         LockEnqueue,
                         LockWait,
                         MscpRead,
                         MscpWrite,
                         PageFaults,
                         PageReadIO,
                         PageWriteIO,
                         ProcessCount,
                         ProcessBatCount,
                         ProcessIntCount,
                         ProcessNetCount,
                         SpiEfn,
                         ScsDataSize;

   static unsigned long  PrevBufferedIO,
                         PrevClusterMsgSent,
                         PrevClusterMsgRcvd,
                         PrevClusterKbyteMap,
                         PrevClusterKbyteSent,
                         PrevClusterKbyteRcvd,
                         PrevDirectIO,
                         PrevFcpRead,
                         PrevFcpWrite,
                         PrevLockActivity,
                         PrevLockWait,
                         PrevMscpRead,
                         PrevMscpWrite,
                         PrevPageFaults,
                         PrevPageReadIO,
                         PrevPageWriteIO,
                         PrevMscpIO;

   static unsigned long  CpuExec [2],
                         CpuIntStk [2],
                         CpuKernel [2],
                         CpuMpSynch [2],
                         CpuSuper [2],
                         CpuUser [2];

   static unsigned long  MscpData [35];

   static unsigned long  PrevCpuTicksBusy [2],
                         PrevCpuTicksUser [2];

   static unsigned long  *ScsDataPtr;

#ifdef __ia64
   static unsigned long  FreeList [2],
                         ModList [2];
#else
   static unsigned long  FreeList,
                         ModList;
#endif

   /* initialize structure for RMI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       /* for 7.3-2(?) and later which has the RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(CpuExec), RMI$_CPUEXEC, &CpuExec, 0 },
       { sizeof(CpuIntStk), RMI$_CPUINTSTK, &CpuIntStk, 0 },
       { sizeof(CpuKernel), RMI$_CPUKERNEL, &CpuKernel, 0 },
       { sizeof(CpuMpSynch), RMI$_CPUMPSYNCH, &CpuMpSynch, 0 },
       { sizeof(CpuSuper), RMI$_CPUSUPER, &CpuSuper, 0 },
       { sizeof(CpuUser), RMI$_CPUUSER, &CpuUser, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(MscpData), RMI$_MSCP_EVERYTHING, &MscpData, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       { sizeof(ProcessIntCount), RMI$_PROCINTCNT, &ProcessIntCount, 0 },
       { sizeof(ProcessBatCount), RMI$_PROCBATCNT, &ProcessBatCount, 0 },
       { sizeof(ProcessNetCount), RMI$_PROCNETCNT, &ProcessNetCount, 0 },
       {0,0,0,0}
   }, ItemList2[] =
   {
       /* for 7.3-1(?) and earlier which do not have RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(ModeTicks), RMI$_MODES, &ModeTicks, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       {0,0,0,0}
   }, ItemList3[] =
   {
       { 0, RMI$_SCS, 0, 0 },
       { sizeof(MscpRead), RMI$_MSCP_READ, &MscpRead, 0 },
       { sizeof(MscpWrite), RMI$_MSCP_WRITE, &MscpWrite, 0 },
       {0,0,0,0}
   }, *ItemListPtr = ItemList1;

   int  cidx, midx, status, tlong;
   unsigned long  datsize;
   unsigned long  CpuTicksBusy [2];
   unsigned long  *datptr;
   float  tfloat;
   IOSBLK  IOsb,
           SynchIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectRMI()\n");

   if (!SpiEfn)
      if (VMSnok (status = lib$get_ef (&SpiEfn)))
         EXIT_FI_LI (status);


   /* collect System Performance Information */
   status = sys$getrmi (
            SpiEfn,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            ItemListPtr,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getrmi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.iosb$w_status);

   /* fall back to using the pre VMS V7.3-2 items */ 
   if (status == SS$_BADPARAM || IOsb.iosb$w_status == SS$_BADPARAM)
   {
      if (ItemListPtr == ItemList1)
      {
         ItemListPtr = ItemList2;
         return (status);
      }
   }

   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
      return (status);
   }

   if (sysptr->ClusterMember)
   {
      /* collect cluster information */
      datsize = ((sysptr->ClusterNodes * 14) + 1) * sizeof(unsigned long);
      if (Debug) fprintf (stdout, "%d %d\n", sysptr->ClusterNodes, datsize);
      if (ScsDataSize < datsize)
      {
         if (ScsDataPtr) free (ScsDataPtr);
         ScsDataPtr = (unsigned long*)calloc (1, ScsDataSize = datsize);
         if (!ScsDataPtr) EXIT_FI_LI (vaxc$errno);
      }
      ItemList3[0].BufferPtr = ScsDataPtr;
      ItemList3[0].BufferLength = ScsDataSize;

      /* does not appear to be supported by $GETRMI */
      status = exe$getspi (
               SpiEfn,  /* efn */
               0,  /* csiaddr */
               0,  /* nodename */
               &ItemList3,  /* item list */
               &IOsb,  /* iosb */
               0,      /* astaddr */
               0 );    /* astprm */

      if (Debug)
         fprintf (stdout, "exe$getspi() %%X%08.08X  IOsb: %%X%08.08X\n",
                  status, IOsb.iosb$w_status);

      if (VMSnok (status))
      {
         sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
         return (status);
      }

      /*** huh? ***/
      ScsDataPtr = ItemList3[0].BufferPtr;

      if (Debug) fprintf (stdout, "%d\n", ScsDataPtr[0]);
      if (ScsDataPtr[0] > sysptr->ClusterNodes)  EXIT_FI_LI (SS$_BUGCHECK);

      for (cidx = 0; cidx < ScsDataPtr[0]; cidx++)
      {
         datptr = &(ScsDataPtr[(cidx*14)+1]);
         if (Debug)
            fprintf (stdout, "%d |%s| %d %d %d %d %d %d %d %d %d %d %d\n",
                     cidx, (char*)(&datptr[0])+1,
                     datptr[2], datptr[3], datptr[4], datptr[5],
                     datptr[6], datptr[7], datptr[8], datptr[9],
                     datptr[10], datptr[11], datptr[12]);
         ClusterMsgSent = datptr[5];
         ClusterMsgRcvd = datptr[6];
         ClusterKbyteSent = datptr[8];
         ClusterKbyteRcvd = datptr[10];
         ClusterKbyteMap = datptr[11];
      }
  }

/***
   status = sys$synch (SpiEfn, &SynchIOsb);

   if (Debug)
      fprintf (stdout, "sys$synch() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, SynchIOsb.iosb$w_status);

   if (VMSnok (status)) return (status);
***/

   /*******/
   /* CPU */
   /*******/

   if (ItemListPtr == ItemList1)
   {
      /* VMS 7.3-2 and later */
      status = lib$addx (&CpuExec, &CpuIntStk, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuKernel, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuMpSynch, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuSuper, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
      status = lib$addx (&CpuUser, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);
   }
   else
   {
     /* VMS V7.3-1 and earlier */
     int  cidx, midx;
     for (cidx = 1; cidx < ModeTicks.NumberOfCPUs; cidx++)
        for (midx = INTSTK_MODE; midx <= IDLE_MODE; midx++)
            ModeTicks.CPU[0].Count[midx] += ModeTicks.CPU[cidx].Count[midx];

     CpuTicksBusy[1] = CpuUser[1] = 0;
     CpuTicksBusy[0] = ModeTicks.CPU[0].Count[INTSTK_MODE];
     /* interrupt item is actually the sum of interrupt and idle! */
     CpuTicksBusy[0] -= ModeTicks.CPU[0].Count[IDLE_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[KERNEL_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[MPSYNCH_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[EXEC_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[SUPER_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[USER_MODE];
     CpuUser[0] = ModeTicks.CPU[0].Count[USER_MODE];

#ifdef  __VAX
     ProcessIntCount = sys$gw_ijobcnt;
     ProcessBatCount = sys$gw_bjobcnt;
     ProcessNetCount = sys$gw_njobcnt;
#else
     ProcessIntCount = sys$gl_ijobcnt;
     ProcessBatCount = sys$gl_bjobcnt;
     ProcessNetCount = sys$gl_njobcnt;
#endif
     /* of course sys$g*_njobcnt doesn't seem to account network jobs! */
     ProcessNetCount = -1;
   }

   if (PrevCpuTicksBusy[0] || PrevCpuTicksBusy[1])
   {
      status = lib$subx (&CpuTicksBusy, &PrevCpuTicksBusy,
                         &sysptr->CpuTicksBusy, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);

      tfloat = ((float)sysptr->CpuTicksBusy[0] +
                ((float)sysptr->CpuTicksBusy[1] * 4294967296.0));
      tfloat /= DeltaSeconds;
      sysptr->CpuBusyPercent = (unsigned int)tfloat;
      if (sysptr->CpuBusyPercent > 100 * sysptr->ActiveCpuCount)
         sysptr->CpuBusyPercent = 100 * sysptr->ActiveCpuCount;
   }
   else
      sysptr->CpuBusyPercent = 0;

   PrevCpuTicksBusy[0] = CpuTicksBusy[0];
   PrevCpuTicksBusy[1] = CpuTicksBusy[1];

   if (PrevCpuTicksUser[0] || PrevCpuTicksUser[1])
   {
      status = lib$subx (&CpuUser, &PrevCpuTicksUser,
                         &sysptr->CpuTicksUser, 0);
      if (VMSnok(status)) EXIT_FI_LI (status);

      tfloat = ((float)sysptr->CpuTicksUser[0] +
                ((float)sysptr->CpuTicksUser[1] * 4294967296.0));
      tfloat /= DeltaSeconds;
      sysptr->CpuUserPercent = (unsigned int)tfloat;
      if (sysptr->CpuUserPercent > 100 * sysptr->ActiveCpuCount)
         sysptr->CpuUserPercent = 100 * sysptr->ActiveCpuCount;
   }
   else
      sysptr->CpuUserPercent = 0;

   PrevCpuTicksUser[0] = CpuUser[0];
   PrevCpuTicksUser[1] = CpuUser[1];

   if (Debug)
      fprintf (stdout, "CPU: %d%% USER: %d%%\n",
               sysptr->CpuBusyPercent, sysptr->CpuUserPercent);

   /***************/
   /* page faults */
   /***************/

   if (PrevPageFaults && PageFaults > PrevPageFaults)
      sysptr->PageFaults = PageFaults - PrevPageFaults;
   else
      sysptr->PageFaults = 0;
   PrevPageFaults = PageFaults;

   if (PrevPageReadIO && PageReadIO > PrevPageReadIO)
      sysptr->PageReadIO = PageReadIO - PrevPageReadIO;
   else
      sysptr->PageReadIO = 0;
   PrevPageReadIO = PageReadIO;

   if (PrevPageWriteIO && PageWriteIO > PrevPageWriteIO)
      sysptr->PageWriteIO = PageWriteIO - PrevPageWriteIO;
   else
      sysptr->PageWriteIO = 0;
   PrevPageWriteIO = PageWriteIO;

   sysptr->PageFaultsHard = sysptr->PageReadIO + sysptr->PageWriteIO;
   tfloat = (float)sysptr->PageFaultsHard / DeltaSeconds;
   sysptr->PageFaultsHard = (unsigned int)tfloat;

   sysptr->PageFaultsSoft = sysptr->PageFaults - sysptr->PageFaultsHard;
   tfloat = (float)sysptr->PageFaultsSoft / DeltaSeconds;
   sysptr->PageFaultsSoft = (unsigned int)tfloat;

   /********/
   /* MSCP */
   /********/

   if (sysptr->ClusterMember)
   {
      if (ItemListPtr == ItemList1)
      {
         MscpRead = MscpData[23];
         MscpWrite = MscpData[24];
      }

      if (PrevMscpRead && MscpRead > PrevMscpRead)
         sysptr->MscpRead = MscpRead - PrevMscpRead;
      else
         sysptr->MscpRead = 0;
      PrevMscpRead = MscpRead;
      tfloat = (float)sysptr->MscpRead / DeltaSeconds;
      sysptr->MscpRead = (unsigned int)tfloat;

      if (PrevMscpWrite > MscpWrite > PrevMscpWrite)
         sysptr->MscpWrite = MscpWrite - PrevMscpWrite;
      else
         sysptr->MscpWrite = 0;
      PrevMscpWrite = MscpWrite;
      tfloat = (float)sysptr->MscpWrite / DeltaSeconds;
      sysptr->MscpWrite = (unsigned int)tfloat;
   }

   /********/
   /* lock */
   /********/

   if (PrevLockActivity &&
       LockEnqueue + LockDequeue + LockWait > PrevLockActivity)
      sysptr->LockActivity = LockEnqueue + LockDequeue + LockWait -
                             PrevLockActivity;
   else
      sysptr->LockActivity = 0;
   PrevLockActivity = LockEnqueue + LockDequeue + LockWait;
   tfloat = (float)sysptr->LockActivity / DeltaSeconds;
   sysptr->LockActivity = (unsigned int)tfloat;
   tfloat = (float)sysptr->LockWait / DeltaSeconds;
   sysptr->LockWait = (unsigned int)tfloat;

   if (PrevLockWait && LockWait > PrevLockWait)
      sysptr->LockWait = LockWait - PrevLockWait;
   else
      sysptr->LockWait = 0;
   PrevLockWait = LockWait;

   /********/
   /* disk */
   /********/

   if (PrevFcpRead && FcpRead > PrevFcpRead)
      sysptr->FcpRead = FcpRead - PrevFcpRead;
   else
      sysptr->FcpRead = 0;
   PrevFcpRead = FcpRead;
   tfloat = (float)sysptr->FcpRead / DeltaSeconds;
   sysptr->FcpRead = (unsigned int)tfloat;

   if (PrevFcpWrite && FcpWrite > PrevFcpWrite)
      sysptr->FcpWrite = FcpWrite - PrevFcpWrite;
   else
      sysptr->FcpWrite = 0;
   PrevFcpWrite = FcpWrite;
   tfloat = (float)sysptr->FcpWrite / DeltaSeconds;
   sysptr->FcpWrite = (unsigned int)tfloat;

   /*********/
   /* other */
   /*********/

   if (PrevDirectIO && DirectIO > PrevDirectIO)
      sysptr->DirectIO = DirectIO - PrevDirectIO;
   else
      sysptr->DirectIO = 0;
   PrevDirectIO = DirectIO;
   tfloat = (float)sysptr->DirectIO / DeltaSeconds;
   sysptr->DirectIO = (unsigned int)tfloat;

   if (PrevBufferedIO && BufferedIO > PrevBufferedIO)
      sysptr->BufferedIO = BufferedIO - PrevBufferedIO;
   else
      sysptr->BufferedIO = 0;
   PrevBufferedIO = BufferedIO;
   tfloat = (float)sysptr->BufferedIO / DeltaSeconds;
   sysptr->BufferedIO = (unsigned int)tfloat;

#ifdef __ia64
   sysptr->MemInUse = sysptr->MemSize - (FreeList[0] >> 7) - (ModList[0] >> 7);
#else
   sysptr->MemInUse = sysptr->MemSize - (FreeList >> 7) - (ModList >> 7);
#endif
   if (sysptr->MemInUse > sysptr->MemInUseMax)
      sysptr->MemInUseMax = sysptr->MemInUse;

   sysptr->ProcessCount = ProcessCount;
   sysptr->ProcessBatCount = ProcessBatCount;
   sysptr->ProcessIntCount = ProcessIntCount;
   sysptr->ProcessNetCount = ProcessNetCount;

   if (sysptr->ClusterMember)
   {
      /*******/
      /* SCS */
      /*******/

      if (PrevClusterMsgSent && ClusterMsgSent > PrevClusterMsgSent)
         sysptr->ClusterMsgSent = ClusterMsgSent - PrevClusterMsgSent;
      else
         sysptr->ClusterMsgSent = 0;
      PrevClusterMsgSent = ClusterMsgSent;
      tfloat = (float)sysptr->ClusterMsgSent / DeltaSeconds;
      sysptr->ClusterMsgSent = (unsigned int)tfloat;

      if (PrevClusterMsgRcvd && ClusterMsgRcvd > PrevClusterMsgRcvd)
         sysptr->ClusterMsgRcvd = ClusterMsgRcvd - PrevClusterMsgRcvd;
      else
         sysptr->ClusterMsgRcvd = 0;
      PrevClusterMsgRcvd = ClusterMsgRcvd;
      tfloat = (float)sysptr->ClusterMsgRcvd / DeltaSeconds;
      sysptr->ClusterMsgRcvd = (unsigned int)tfloat;

      if (PrevClusterKbyteMap && ClusterKbyteMap > PrevClusterKbyteMap)
         sysptr->ClusterKbyteMap = ClusterKbyteMap - PrevClusterKbyteMap;
      else
         sysptr->ClusterKbyteMap = 0;
      PrevClusterKbyteMap = ClusterKbyteMap;
      tfloat = (float)sysptr->ClusterKbyteMap / DeltaSeconds;
      sysptr->ClusterKbyteMap = (unsigned int)tfloat;

      if (PrevClusterKbyteSent && ClusterKbyteSent > PrevClusterKbyteSent)
         sysptr->ClusterKbyteSent = ClusterKbyteSent - PrevClusterKbyteSent;
      else
         sysptr->ClusterKbyteSent = 0;
      PrevClusterKbyteSent = ClusterKbyteSent;
      tfloat = (float)sysptr->ClusterKbyteSent / DeltaSeconds;
      sysptr->ClusterKbyteSent = (unsigned int)tfloat;

      if (PrevClusterKbyteRcvd && ClusterKbyteRcvd > PrevClusterKbyteRcvd)
         sysptr->ClusterKbyteRcvd = ClusterKbyteRcvd - PrevClusterKbyteRcvd;
      else
         sysptr->ClusterKbyteRcvd = 0;
      PrevClusterKbyteRcvd = ClusterKbyteRcvd;
      tfloat = (float)sysptr->ClusterKbyteRcvd / DeltaSeconds;
      sysptr->ClusterKbyteRcvd = (unsigned int)tfloat;
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function uses undocumented sense-mode functionality to load counters from
the network interface (commonly ethernet device).  Multiple devices are
supported using a multi-valued logical name MONDESI_NI and the frame-rate
and byte-rate and soft/hard error-rate statistics are accumlated to represent
total system data.

On Alpha and Itanium uses native 64 bit (quadword) accumulators.

Essential concepts have come from "EMON: moniteur ethernet (V2.1-2) Gerard
Guillaume" and found in DECUSLIB freeware, and from some peeks at the T4 source
code.  The implementation is mine.
*/

int CollectNetInt (SYSTEM_DATA *sysptr)

{
   static int  NetIntCount;
   static unsigned short  NetIntChan [NI_DEVICE_MAX];
   static unsigned long  DcSCOM = 32,  /* DC$_SCOM */
                         DtNI = 13;    /* DT$_NI */

#ifndef __VAX
   static unsigned __int64  PrevBlocksRx,
                            PrevBlocksTx,
                            PrevBytesRx,
                            PrevBytesTx,
                            PrevErrorsHard,
                            PrevErrorsSoft;
#else
   static unsigned long  PrevBlocksRx,
                         PrevBlocksTx,
                         PrevBytesRx,
                         PrevBytesTx,
                         PrevErrorsHard,
                         PrevErrorsSoft;
#endif

   static long  DeviceCount = -1,
                DviUnit,
                LnmIndex = -1,
                LnmAttributes;
   static unsigned short  LogValueLength;
   static char  LogValue [32];
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { sizeof(LogValue)-1, LNM$_STRING, &LogValue, &LogValueLength },
      { 0,0,0,0 }
   },
   DevScanItems [] =
   {
      { sizeof(DcSCOM), DVS$_DEVCLASS, &DcSCOM, 0 },
      { 0,0,0,0 }
   },
   GetDviItems [] =
   {
      { sizeof(DviUnit), DVI$_UNIT, &DviUnit, 0 },
      { 0,0,0,0 }
   };
   static $DESCRIPTOR (NetIntDsc, "");
   static $DESCRIPTOR (LogNameDsc, "MONDESI_NI");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");

   int  ecnt, status;
   unsigned short  pword,
                   retlen;
   unsigned long  LnmCount,
                  NetIntErrorsHard,
                  NetIntErrorsSoft;
   unsigned long  DevScanContext [2];
   float  tfloat;
   char  *vptr;
   char  DevScanBuffer [64],
         CountBuffer [512];
   struct STRUCT_IOSBLK  IOsb;
   $DESCRIPTOR (CountDsc, CountBuffer);
   $DESCRIPTOR (DevScanDsc, DevScanBuffer);

#ifndef __VAX
#define CASTIT (unsigned __int64)
   unsigned __int64  value,
                     NetIntBlocksRx,
                     NetIntBlocksTx,
                     NetIntBytesRx,
                     NetIntBytesTx;
#else
#define CASTIT (unsigned long)
   unsigned long  value,
                  NetIntBlocksRx,
                  NetIntBlocksTx,
                  NetIntBytesRx,
                  NetIntBytesTx;
#endif

   /*********/
   /* begin */
   /*********/

   if (LnmIndex < 0)
   {
      /* first call, check for logical name defined network devices */
      for (LnmIndex = 0; LnmIndex < NI_DEVICE_MAX; LnmIndex++)
      {
         status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, 0, &LnmItems);
         if (VMSnok (status) || !(LnmAttributes & LNM$M_EXISTS)) break;
         LogValue[LogValueLength] = '\0';

         NetIntDsc.dsc$a_pointer = LogValue;
         NetIntDsc.dsc$w_length = LogValueLength;

         status = sys$assign (&NetIntDsc, &NetIntChan[LnmIndex], 0, 0);
         if (VMSnok (status))
         {
            sprintf (NotePad, "$[%d] %s %%X%08.08X",
                     __LINE__, LogValue, status);
            NetIntCount = 0;
            break;
         }
         if (NetIntDev[0]) strcat (NetIntDev, ",");
         strcat (NetIntDev, LogValue);
         NetIntCount++;
      }
   }

   if (DeviceCount < 0)
   {
      /* first call */
      DeviceCount = 0;

      /* if logical name defined and found devices */
      if (NetIntCount) return (NetIntCount);

      /* scan system for network devices */
      DevScanContext[0] = DevScanContext[1] = 0;
      while (NetIntCount < NI_DEVICE_MAX)
      {
         status = sys$device_scan (&DevScanDsc, &retlen, 0,
                                   &DevScanItems, &DevScanContext);
         if (VMSnok (status))
         {
            if (status == SS$_NOMOREDEV || status == SS$_NOSUCHDEV) break;
            sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
            NetIntCount = 0;
            break;
         }

         status = sys$getdviw (0, 0, &DevScanDsc, &GetDviItems, &IOsb, 
                               0, 0, 0, 0);
         if (VMSok (status)) status = IOsb.iosb$w_status;
         if (VMSnok (status))
         {
            sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
            NetIntCount = 0;
            break;
         }
         /* if not a template device then continue */
         if (DviUnit != 0) continue;

         status = sys$assign (&DevScanDsc, &NetIntChan[DeviceCount++], 0, 0);
         if (VMSnok (status))
         {
            if (status == SS$_NOPRIV) continue;
            sprintf (NotePad, "$[%d] %s %%X%08.08X",
                     __LINE__, LogValue, status);
            NetIntCount = 0;
            break;
         }

         if (NetIntDev[0]) strcat (NetIntDev, ",");
         DevScanBuffer[retlen] = '\0';
         strcat (NetIntDev, DevScanBuffer+1);
         NetIntCount++;
      }

      /* first call, just return having assigned channels (perhaps) */
      return (NetIntCount);
   }

   /****************/
   /* collect data */
   /****************/

   /* initialise structure and local storage */
   sysptr->NetIntBlocksRx = sysptr->NetIntBlocksTx =
      sysptr->NetIntBytesRx = sysptr->NetIntBytesTx =
      NetIntBlocksRx = NetIntBlocksTx =
      NetIntBytesRx = NetIntBytesTx =
      NetIntErrorsHard = NetIntErrorsSoft = 0;

   /* if no devices then just return from here */
   if (!NetIntCount) return (NetIntCount);

   /* accumulate counts from each device */
   for (ecnt = 0; ecnt < NetIntCount; ecnt++)
   {
      memset (CountBuffer, 0, sizeof(CountBuffer));

      status = sys$qiow (0, NetIntChan[ecnt],
#ifndef __VAX
#define IO$M_RD_64COUNT 0x4000
                         IO$_SENSEMODE | IO$M_RD_COUNT |
                         IO$M_RD_64COUNT | IO$M_CTRL,
#else
                         IO$_SENSEMODE | IO$M_RD_COUNT | IO$M_CTRL,
#endif
                         &IOsb, 0, 0,
                         0, &CountDsc, 0, 0, 0, 0);

      if (VMSok (status)) status = IOsb.iosb$w_status;
      if (VMSnok (status))
      {
         sprintf (NotePad, "$[%d] %%X%08.08X", __LINE__, status);
         return (NetIntCount);
      }

      vptr = CountBuffer;
      while (vptr < CountBuffer + IOsb.iosb$w_bcnt)
      {
         pword = *(unsigned short*)vptr;
         vptr += 2;

         /* end of list */
         if (!pword) break;

         /* MSB should be set! */
         if (!(pword & 0x8000))
         {
            sprintf (NotePad, "$[%d] %04.04X", __LINE__, pword);
            break;
         }

         value = 0;
#ifndef __VAX
         if ((pword & 0x0fff) < 200)
         {
            /* quadword value (item number less than 200) */
            value = *(unsigned __int64*)vptr;
            vptr += 8;
         }
         else
#endif
         if ((pword & 0x6000) == 0x6000)
         {
            /* longword value */
            value = *(unsigned long*)vptr;
            vptr += 4;
         } 
         else
         if ((pword & 0x6000) == 0x4000)
         {
            /* word value */
            value = *(unsigned short*)vptr;
            vptr += 2;
         } 
         else
         if ((pword & 0x6000) == 0x2000)
         {
            /* byte value */
            value = *(unsigned char*)vptr;
            vptr++;
         } 

         if (pword & 0x1000)
         {
            /* hw map (?) */
            value = *(unsigned long*)vptr;
            vptr += 2;
            continue;
         } 

         switch (pword & 0x0fff)
         {
#ifndef __VAX
            /* these look suspiciously like "LANCP SHOW DEVICE /COUNT"! */

            case 1 :
                 /* seconds since last zeroed */
                 break;

            case 2 :
            case 6 :
                 NetIntBytesRx += value;
                 break;

            case 3 :
            case 7 :
                 NetIntBytesTx += value;
                 break;

            case 4 :
            case 8 :
                 NetIntBlocksRx += value;
                 break;

            case 5 :
            case 9 :
                 NetIntBlocksTx += value;
                 break;

            case 12 :
            case 13 :
            case 14 :
            case 15 :
            case 16 :
            case 17 :
            case 18 :
            case 19 :
            case 30 :
            case 21 :
            case 22 :
            case 23 :
            case 24 :
            case 27 :
            case 28 :
            case 29 :
                 NetIntErrorsSoft += value;
                 break;

            case 25 :
            case 26 :
            case 33 :
            case 34 :
                 NetIntErrorsHard += value;
                 break;
#endif

            /* more "traditional" counters */

            case 1000 /* NMA$C_CTLIN_BRC */ :
            case 1002 /* NMA$C_CTLIN_MBY */ :
                 NetIntBytesRx += value;
                 break;

            case 1001 /* NMA$C_CTLIN_BSN */ :
                 NetIntBytesTx += value;
                 break;

            case 1010 /* NMA$C_CTLIN_DBR */ :
            case 1012 /* NMA$C_CTLIN_MBL */ :
                 NetIntBlocksRx += value;
                 break;

            case 1011 /* NMA$C_CTLIN_DBS */ :
                 NetIntBlocksTx += value;
                 break;

            case 1013 /* NMA$C_CTLIN_BID */ :
            case 1014 /* NMA$C_CTLIN_BS1 */ :
            case 1015 /* NMA$C_CTLIN_BSM */ :
            case 1040 /* NMA$C_CTLIN_RBE */ :
            case 1041 /* NMA$C_CTLIN_LBE */ :
            case 1064 /* NMA$C_CTLIN_OVR */ :
            case 1066 /* NMA$C_CTLIN_UBU */ :

                 NetIntErrorsSoft += value;
                 break;

            case 1060 /* NMA$C_CTLIN_SFL */ :
            case 1061 /* NMA$C_CTLIN_CDC */ :
            case 1063 /* NMA$C_CTLIN_UFD */ :
            case 1062 /* NMA$C_CTLIN_RFL */ :
            case 1065 /* NMA$C_CTLIN_SBU */ :

                 NetIntErrorsHard += value;
                 break;

            default :
         }
      }
   }

   if (PrevBytesRx && NetIntBytesRx > PrevBytesRx)
      sysptr->NetIntBytesRx = NetIntBytesRx - PrevBytesRx;
   PrevBytesRx = NetIntBytesRx;

   if (PrevBytesTx && NetIntBytesTx > PrevBytesTx)
      sysptr->NetIntBytesTx = NetIntBytesTx - PrevBytesTx;
   PrevBytesTx = NetIntBytesTx;

   tfloat = (float)sysptr->NetIntBytesRx / DeltaSeconds;
   sysptr->NetIntBytesRx = CASTIT tfloat;

   tfloat = (float)sysptr->NetIntBytesTx / DeltaSeconds;
   sysptr->NetIntBytesTx = CASTIT tfloat;

   if (PrevBlocksRx && NetIntBlocksRx > PrevBlocksRx)
      sysptr->NetIntBlocksRx = NetIntBlocksRx - PrevBlocksRx;
   PrevBlocksRx = NetIntBlocksRx;

   if (PrevBlocksTx && NetIntBlocksTx > PrevBlocksTx)
      sysptr->NetIntBlocksTx = NetIntBlocksTx - PrevBlocksTx;
   PrevBlocksTx = NetIntBlocksTx;

   tfloat = (float)sysptr->NetIntBlocksRx / DeltaSeconds;
   sysptr->NetIntBlocksRx = CASTIT tfloat;

   tfloat = (float)sysptr->NetIntBlocksTx / DeltaSeconds;
   sysptr->NetIntBlocksTx = CASTIT tfloat;

   if (PrevErrorsSoft && NetIntErrorsSoft > PrevErrorsSoft)
      sysptr->NetIntErrorsSoft += NetIntErrorsSoft - PrevErrorsSoft;
   PrevErrorsSoft = NetIntErrorsSoft;

   if (PrevErrorsHard && NetIntErrorsHard > PrevErrorsHard)
      sysptr->NetIntErrorsHard += NetIntErrorsHard - PrevErrorsHard;
   PrevErrorsHard = NetIntErrorsHard;

   return (NetIntCount);
}

/*****************************************************************************/
/*
Sets up a WebSocket client which then runs independently and asynchronously in
a CGIplus environment.
*/

void WebSockRequest (int RefreshInterval)

{
   char  *cptr;
   CLIENT_DATA  *clptr;

   /*********/
   /* begin */
   /*********/

   /* report to the server that the WebSocket client is accepted */
   fprintf (stdout, "HTTP/1.1 101 Switching Protocols\r\n\r\n");
   fflush (stdout);

   /* allocate a data structure for this connected client */
   clptr = calloc (1, sizeof(CLIENT_DATA));
   if (!clptr) EXIT_FI_LI (vaxc$errno);

   clptr->RefreshSeconds = RefreshInterval;

   /* open the WebSocket IPC to the server (and from there to the client) */
   WebSockOpen (clptr);

   /* begin updating the data at the client end */
   WebSockUpdate (clptr);

   /* begin asynchronous reads from client */
   WebSockRead (clptr);

   /* standard CGI waits for WebSocket to close */
   if (!IsCgiPlus) WebSockWait ();
}

/*****************************************************************************/
/*
Wait for the/a WebSocket to close (and request to conclude).
*/

void WebSockWait ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockWait()\n");

   while (ConnectedCount) sys$hiber ();
}

/*****************************************************************************/
/*
Timer AST driven update of the WebSocket client.  One to sixty seconds.
*/

void WebSockUpdate (CLIENT_DATA *clptr)

{
   int  len, status;
   int  SecondsDelta [2] = { -10000000, -1 };


   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockUpdate()\n");

   len = MonDeSiPrepareData (clptr);

   if (len) WebSockWrite (clptr, clptr->OutputBuffer, len);

   /* one to sixty seconds, just multiply by interval */
   SecondsDelta[0] *= clptr->RefreshSeconds;

   status = sys$setimr (0, &SecondsDelta, WebSockUpdate, clptr, 0);

   if (VMSnok(status)) EXIT_FI_LI (status);
}

/*****************************************************************************/
/*
Using the device names from the CGI variables WEBSOCKET_INPUT and
WEBSOCKET_OUTPUT assign channels in preparation for asynchronous I/O.
*/

void WebSockOpen (CLIENT_DATA *clptr)

{
   int  status;
   char  *cptr;
   $DESCRIPTOR (MbxDsc, "");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockOpen()\n");

   cptr = GetCgiVar ("WEBSOCKET_INPUT");
   if (cptr == NULL) EXIT_FI_LI (SS$_BUGCHECK);
   strcpy (clptr->InputDevName, cptr);
   MbxDsc.dsc$a_pointer = cptr;
   MbxDsc.dsc$w_length = strlen(cptr);

   status = sys$assign (&MbxDsc, &clptr->InputChannel, 0, 0);
   if (VMSnok(status)) EXIT_FI_LI (status);

   cptr = GetCgiVar ("WEBSOCKET_OUTPUT");
   if (cptr == NULL) EXIT_FI_LI (SS$_BUGCHECK);
   strcpy (clptr->OutputDevName, cptr);
   MbxDsc.dsc$a_pointer = cptr;
   MbxDsc.dsc$w_length = strlen(cptr);

   status = sys$assign (&MbxDsc, &clptr->OutputChannel, 0, 0);
   if (VMSnok(status)) EXIT_FI_LI (status);

   ConnectedCount++;

   /* cancel any existing exit */
   ExitTimeStamp = 0;
}

/*****************************************************************************/
/*
Cancel any outstanding I/O and deassign channels.
*/

void WebSockClose (CLIENT_DATA *clptr)

{

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockClose()\n");

   if (clptr->InputChannel)
   {
      if (clptr->QueuedInput)
         sys$cancel (clptr->InputChannel);
      else
      {
         sys$dassgn (clptr->InputChannel);
         clptr->InputChannel = 0;
      }
   }

   if (clptr->OutputChannel)
   {
      if (clptr->QueuedOutput)
         sys$cancel (clptr->OutputChannel);
      else
      {
         /* let the server know the client is being disconnected */
         sys$qio (0, clptr->OutputChannel,
                  IO$_WRITEOF | IO$M_NORSWAIT, 0, 0, 0,
                  0, 0, 0, 0, 0, 0);
         sys$dassgn (clptr->OutputChannel);
         clptr->OutputChannel = 0;
      }
   }

   if (clptr->InputChannel || clptr->OutputChannel) return;

   /* cancel any update timer */
   sys$cantim (clptr, 0); 

   /* dispose of client */
   free (clptr);

   if (ConnectedCount) ConnectedCount--;

   /* if not longer any connected clients set up an exit strategy */
   if (!ConnectedCount) ExitTimeStamp = TimeStamp + EXIT_AFTER_SECONDS;
}

/*****************************************************************************/
/*
Queue an asynchronous read of the client WEBSOCKET_INPUT mailbox.
*/

void WebSockRead (CLIENT_DATA *clptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockRead()\n");

   status = sys$qio (0, clptr->InputChannel, IO$_READLBLK,
                     &clptr->InputIOsb, WebSockReadAst, clptr,
                     clptr->InputBuffer, sizeof(clptr->InputBuffer)-1,
                     0, 0, 0, 0);

   clptr->QueuedInput++;
}

/*****************************************************************************/
/*
Called when an asynchronous read of the client WEBSOCKET_INPUT completes.
This function drives the client-end business logic.

For MONDESI this is really just a demonstration of the WebSocket asynchronous
client to server communication, where the cluster statistics are turned on or
off (saving a small amount of bandwidth) depending on whether they are visible
or not.
*/

void WebSockReadAst (CLIENT_DATA *clptr)

{
   int  RefreshInterval;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "WebSockReadAst() %%X%08.08X\n",
               clptr->InputIOsb.iosb$w_status);

   if (clptr->QueuedInput) clptr->QueuedInput--;

   if (VMSnok (clptr->InputIOsb.iosb$w_status))
   {
      WebSockClose (clptr);
      return;
   }

   if (clptr->InputIOsb.iosb$w_bcnt >= 2 &&
       (unsigned char)clptr->InputBuffer[0] == 0xff &&
       (unsigned char)clptr->InputBuffer[1] == 0x00)
   {
      /* peer close frame */
      WebSockClose (clptr);
      return;
   }

   /* null-terminate and if necessary remove Web Socket framing */
   clptr->InputBuffer[clptr->InputIOsb.iosb$w_bcnt] = '\0';
   if (WebSockUnFrame (clptr->InputBuffer) == -1)
   {
      /* error in the framing */
      WebSockClose (clptr);
      return;
   }

   if (!strncmp (clptr->InputBuffer, "cluster=", 8))
      clptr->ReportCluster = atoi(clptr->InputBuffer+8);
   else
      sprintf (NotePad, "$[%d] |%16.16s| %04.04x%04.04x%04.04x",
               __LINE__, clptr->InputBuffer,
                         ntohl(*((unsigned long*)clptr->InputBuffer)),
                         ntohl(*((unsigned long*)(clptr->InputBuffer+4))),
                         ntohl(*((unsigned long*)(clptr->InputBuffer+8))));

   WebSockRead (clptr);
}

/*****************************************************************************/
/*
Asynchronous output to client.
*/

void WebSockWrite
(
CLIENT_DATA *clptr,
char *DataPtr,
int DataCount
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WebSockWrite()\n");

   clptr->QueuedOutput++;

   status = sys$qio (0, clptr->OutputChannel, IO$_WRITELBLK,
                     &clptr->OutputIOsb, WebSockWriteAst, clptr,
                     DataPtr, DataCount, 0, 0, 0, 0);
}

/*****************************************************************************/
/*
Output to client WEBSOCKET_OUTPUT has completed.
*/

void WebSockWriteAst (CLIENT_DATA *clptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "WebSockWriteAst() %%X%08.08X\n",
               clptr->OutputIOsb.iosb$w_status);

   if (clptr->QueuedOutput) clptr->QueuedOutput--;

   if (VMSnok (clptr->OutputIOsb.iosb$w_status)) WebSockClose (clptr);
}

/*****************************************************************************/
/*
Remove the Web Socket UTF-8 frame bytes, null-terminate the resultant string,
and return the redacted length, or -1 for an error.  All characters transmitted
with this application will be ASCII (7 bits) so there is no need for UTF-8
decode, return an error if the MSB is set!
*/

int WebSockUnFrame (char *FrameBuffer)

{
   unsigned char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   cptr = sptr = (unsigned char*)FrameBuffer; 
   if (*cptr) return (-1);
   for (cptr++; *cptr && *cptr != 0xff; *sptr++ = *cptr++)
      if (*cptr & 0x80) return (-1);
   *sptr = '\0';
   return (sptr-(unsigned char*)FrameBuffer);
}

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!  Returns the size of the resultant string.
*/

int UrlDecode (char *String)

{
   unsigned char  ch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            ch = 0;
            if (*cptr >= '0' && *cptr <= '9')
               ch = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               ch += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = ch;

            continue;

         case '\r' :
         case '\n' :

            cptr++;
            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

/*****************************************************************************/
/*
Generic, SELF-CONTAINED FUNCTION to return the values of a CGI variable
regardless of whether it is used in a standard CGI environment or a WASD
CGIplus environment.  Also automatically switches WASD V7.2 and later servers
into 'struct' mode for significantly improved performance.  To "force" the use
of 'record' mode (perhaps for performance comparison purposes) assign a
non-empty environment variable (symbol or logical) named CGIPLUS_VAR_RECORD.

Call with 'VarName' empty ("") to synchronize CGIplus requests.  This waits for
a CGIplus variable stream, checks if it's in 'record' or 'struct' mode, reads
the stream appropriately before returning ready to provide variables.

DO NOT modify the character string returned by this function.  Copy to other
storage if this is necessary.  The behaviour is indeterminate if the returned
values are modified in any way!

All CGIplus variables may be returned by making successive calls using a
'VarName' of "*" (often useful when debugging).  NULL is returned when
variables exhausted.

Required standard library header files: stdlib.h, stdio.h, string.h
Required function prototypes: char* GetCgiVar (char*);
                              char* GetCgiVarNull (char*);
*/
/*****************************************************************************/

char* GetCgiVar (char* VarName)
{
   char  *cptr;
   if (cptr = GetCgiVarNull (VarName))
      return (cptr);
   else
   if (VarName && VarName[0] == '*')
      return (cptr);
   else
      return ("");
}

char* GetCgiVarNull (char *VarName)
{
#define CGIVARDEBUG 0

#  ifndef CGIVAR_STRUCT_SIZE
#     define CGIVAR_STRUCT_SIZE 8192
#  endif
#  define SOUS sizeof(unsigned short)

   static int  CalloutDone,
               StructLength,
               WwwPrefix;
   static char  *CgiPlusVarRecordPtr,
                *IsCgiPlusPtr,
                *NextVarNamePtr;
   static char  StructBuffer [CGIVAR_STRUCT_SIZE];
   static FILE  *CgiPlusIn;

   int  Length;
   char  WwwVarName [256];
   char  *bptr, *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (CGIVARDEBUG)
      if (VarName != NULL && VarName[0])
         fprintf (stdout, "CgiVar() |%s|\n",
                  VarName == NULL ? "NULL" : VarName);

   if (VarName == NULL || !VarName[0])
   {
      /* initialize */
      StructLength = WwwPrefix = 0;
      NextVarNamePtr = StructBuffer;
      if (VarName == NULL) return (NULL);
   }

   /* detect the CGIplus environment (once) */
   if (IsCgiPlusPtr == NULL)
   {
      if ((IsCgiPlusPtr = getenv ("CGIPLUSEOF")) == NULL)
      {
         IsCgiPlusPtr = "";
         if (getenv ("WWW_SERVER_SOFTWARE") != NULL) WwwPrefix = 1;
      }
   }

   if (VarName[0])
   {
      /***************************/
      /* return a variable value */
      /***************************/

      if (*(unsigned long*)VarName == 'WWW_' && !WwwPrefix)
         VarName += 4;
      else
      if (*(unsigned long*)VarName != 'WWW_' && WwwPrefix)
      {
         strcpy (WwwVarName, "WWW_");
         strcpy (WwwVarName+4, VarName);
         VarName = WwwVarName;
      }

      if (!IsCgiPlusPtr[0])
      {
         /* standard CGI environment */
         if ((cptr = getenv (VarName)) == NULL) cptr = NULL;
         if (CGIVARDEBUG)
            fprintf (stdout, "CGI |%s=%s|\n",
                     VarName, cptr == NULL ? "NULL" : cptr);
         return (cptr);
      }

      /* hmmm, not initialized */
      if (!StructLength) return (NULL);

      if (VarName[0] == '*')
      {
         /* return each CGIplus variable in successive calls */
         if (!(Length = *(unsigned short*)NextVarNamePtr))
         {
            NextVarNamePtr = StructBuffer;
            return (NULL);
         }
         sptr = (NextVarNamePtr += SOUS);
         NextVarNamePtr += Length;
         return (sptr);
      }

      /* return a pointer to this CGIplus variable's value */
      for (bptr = StructBuffer; Length = *(unsigned short*)bptr; bptr += Length)
      {
         sptr = (bptr += SOUS);
         for (cptr = VarName; *cptr && *sptr && *sptr != '='; cptr++, sptr++)
            if (toupper(*cptr) != toupper(*sptr)) break;
         /* if found return a pointer to the value */
         if (!*cptr && *sptr == '=')
         {
            if (CGIVARDEBUG) fprintf (stdout, "CGIplus |%s|\n", sptr+1);
            return (sptr+1);
         }
      }
      /* not found */
      if (CGIVARDEBUG)
         fprintf (stdout, "CGIplus |%s|\n",
                  NULL == NULL ? "NULL" : NULL);
      return (NULL);
   }

   /*****************************/
   /* get the CGIplus variables */
   /*****************************/

   /* cannot "sync" in a non-CGIplus environment */
   if (!VarName[0] && !IsCgiPlusPtr[0]) return (NULL);

   WwwPrefix = 0;

   /* the CGIPLUSIN stream can be left open */
   if (CgiPlusIn == NULL)
      if ((CgiPlusIn = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
         EXIT_FI_LI (vaxc$errno);

   /* get the starting record (the essentially discardable one) */
   for (;;)
   {
      cptr = fgets (StructBuffer, sizeof(StructBuffer), CgiPlusIn);
      if (cptr == NULL) EXIT_FI_LI (vaxc$errno);
      /* if the starting sentinal is detected then break */
      if (*(unsigned short*)cptr == '!\0' ||
          *(unsigned short*)cptr == '!\n' ||
          (*(unsigned short*)cptr == '!!' && isdigit(*(cptr+2)))) break;
   }

   /* MUST be done after reading the synchronizing starting record */
   if (CGIVARDEBUG) fprintf (stdout, "Content-Type: text/plain\n\n");

   /* detect the CGIplus "force" record-mode environment variable (once) */
   if (CgiPlusVarRecordPtr == NULL)
      if ((CgiPlusVarRecordPtr = getenv ("CGIPLUS_VAR_RECORD")) == NULL)
         CgiPlusVarRecordPtr = "";

   if (*(unsigned short*)cptr == '!!' && !CgiPlusVarRecordPtr[0])
   {
      /********************/
      /* CGIplus 'struct' */
      /********************/

      /* get the size of the binary structure */
      StructLength = atoi(cptr+2);
      if (StructLength <= 0 || StructLength > sizeof(StructBuffer))
         EXIT_FI_LI (SS$_BUGCHECK);

      if (!fread (StructBuffer, 1, StructLength, CgiPlusIn))
         EXIT_FI_LI (vaxc$errno);
   }
   else
   {
      /*********************/
      /* CGIplus 'records' */
      /*********************/

      /* reconstructs the original 'struct'ure from the records */
      sptr = (bptr = StructBuffer) + sizeof(StructBuffer);
      while (fgets (bptr+SOUS, sptr-(bptr+SOUS), CgiPlusIn) != NULL)
      {
         /* first empty record (line) terminates variables */
         if (bptr[SOUS] == '\n') break;
         /* note the location of the length word */
         cptr = bptr;
         for (bptr += SOUS; *bptr && *bptr != '\n'; bptr++);
         if (*bptr != '\n') EXIT_FI_LI (SS$_BUGCHECK);
         *bptr++ = '\0';
         if (bptr >= sptr) EXIT_FI_LI (SS$_BUGCHECK);
         /* update the length word */
         *(unsigned short*)cptr = bptr - (cptr + SOUS);
      }
      if (bptr >= sptr) EXIT_FI_LI (SS$_BUGCHECK);
      /* terminate with a zero-length entry */
      *(unsigned short*)bptr = 0;
      StructLength = (bptr + SOUS) - StructBuffer;
   }

   if (CGIVARDEBUG)
   {
      fprintf (stdout, "%d\n", StructLength);
      for (bptr = StructBuffer; Length = *(unsigned short*)bptr; bptr += Length)
         fprintf (stdout, "|%s|\n", bptr += SOUS);
   }

   if (!CalloutDone && !CgiPlusVarRecordPtr[0])
   {
      /* provide the CGI callout to set CGIplus into 'struct' mode */
      fflush (stdout);
      fputs (CgiPlusEscPtr, stdout);
      fflush (stdout);
      /* the leading '!' indicates we're not going to read the response */
      fputs ("!CGIPLUS: struct", stdout);
      fflush (stdout);
      fputs (CgiPlusEotPtr, stdout);
      fflush (stdout);
      /* don't need to do this again (the '!!' tells us what mode) */
      CalloutDone = 1;
   }

   sptr = StructBuffer + SOUS;
   if (*(unsigned long*)sptr == 'WWW_') WwwPrefix = 1;

   return (NULL);

#  undef SOUS
}

/****************************************************************************/

